# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging
from unittest import TestCase
from hamcrest import assert_that, equal_to, none, instance_of, only_contains, \
    raises

from storops.exception import UnityException, \
    UnityOneSmbServerPerNasServerError, UnityResourceNotFoundError, \
    UnityNetBiosNameExistedError
from storops.unity.resource.cifs_server import UnityCifsServer, \
    UnityCifsServerList
from storops.unity.resource.health import UnityHealth
from storops.unity.resource.interface import UnityFileInterfaceList
from storops.unity.resource.nas_server import UnityNasServer
from storops_test.unity.rest_mock import t_rest, patch_rest

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


class UnityCifsServerTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        server = UnityCifsServer('cifs_2', t_rest())
        assert_that(server.existed, equal_to(True))
        assert_that(server.name, equal_to('nas1130'))
        assert_that(server.description, none())
        assert_that(server.netbios_name, equal_to('NAS1130'))
        assert_that(server.domain, equal_to('win2012.dev'))
        assert_that(server.last_used_organizational_unit,
                    equal_to('ou=Computers,ou=EMC VNX'))
        assert_that(server.workgroup, none())
        assert_that(server.is_standalone, equal_to(False))
        assert_that(server.health, instance_of(UnityHealth))
        assert_that(server.nas_server, instance_of(UnityNasServer))
        assert_that(server.file_interfaces,
                    instance_of(UnityFileInterfaceList))
        assert_that(len(server.file_interfaces), equal_to(1))
        assert_that(server.smbca_supported, equal_to(True))
        assert_that(server.smb_multi_channel_supported, equal_to(True))
        assert_that(server.smb_protocol_versions,
                    only_contains('1.0', '2.0', '2.1', '3.0'))

    @patch_rest
    def test_get_all(self):
        servers = UnityCifsServerList(cli=t_rest())
        assert_that(len(servers), equal_to(1))

    @patch_rest
    def test_create_domain_not_specified(self):
        def f():
            UnityCifsServer.create(t_rest(), 'nas_2', name='c_server1')

        assert_that(f, raises(UnityException, 'domain has not been specified'))

    @patch_rest
    def test_create_password_criteria(self):
        def f():
            UnityCifsServer.create(t_rest(), 'nas_2', name='c_server1',
                                   workgroup='CEDRIC',
                                   local_password='password')

        assert_that(f, raises(UnityException, 'not meet the password policy'))

    @patch_rest
    def test_create_one_smb_server_allowed(self):
        def f():
            UnityCifsServer.create(t_rest(), 'nas_2', name='c_server1',
                                   workgroup='CEDRIC',
                                   local_password='Password123!')

        assert_that(f, raises(UnityOneSmbServerPerNasServerError,
                              'Only one SMB server'))

    @patch_rest
    def test_create_success(self):
        server = UnityCifsServer.create(t_rest(), 'nas_5', name='c_server1',
                                        workgroup='CEDRIC',
                                        local_password='Password123!')
        assert_that(server.workgroup, equal_to('CEDRIC'))

    @patch_rest
    def test_create_name_existed(self):
        def f():
            UnityCifsServer.create(t_rest(), 'nas_5', name='NAS1130',
                                   workgroup='CEDRIC',
                                   local_password='Password123!')

        assert_that(f, raises(UnityNetBiosNameExistedError, 'already exists'))

    @patch_rest
    def test_delete_cifs3(self):
        server = UnityCifsServer(_id='cifs_3', cli=t_rest())
        resp = server.delete(skip_domain_unjoin=True)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_delete_not_found(self):
        def f():
            server = UnityCifsServer(_id='cifs_10', cli=t_rest())
            server.delete()

        assert_that(f, raises(UnityResourceNotFoundError))

    @patch_rest
    def test_create_cifs_share_success(self):
        server = UnityCifsServer.get(_id='cifs_2', cli=t_rest())
        share = server.create_cifs_share('cs1', 'fs_8')
        assert_that(share.name, equal_to('cs1'))
        assert_that(share.existed, equal_to(True))

    @patch_rest
    def test_get_cifs_server_from_nas_server(self):
        server = UnityNasServer(_id='nas_2', cli=t_rest())
        server = UnityCifsServer.get(t_rest(), server)
        assert_that(server, instance_of(UnityCifsServer))
        assert_that(server.domain, equal_to('win2012.dev'))

    @patch_rest
    def test_get_from_id(self):
        server = UnityCifsServer.get(cli=t_rest(), _id='cifs_2')
        assert_that(server, instance_of(UnityCifsServer))
        assert_that(server.domain, equal_to('win2012.dev'))

    @patch_rest
    def test_get_from_cifs_server(self):
        cifs_2 = UnityCifsServer(_id='cifs_2', cli=t_rest())
        server = UnityCifsServer.get(cli=t_rest(), _id=cifs_2)
        assert_that(server.domain, equal_to('win2012.dev'))

    @patch_rest
    def test_cifs_server(self):
        server = UnityCifsServer(_id='cifs_2', cli=t_rest())
        log.debug('netbios name: {}'.format(server.netbios_name))
        return server



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from retryz import retry

import storops.vnx.resource.block_pool
import storops.vnx.resource.cg
import storops.vnx.resource.mirror_view
import storops.vnx.resource.sg
from storops import exception as ex
from storops.lib.common import check_int, daemon, instance_cache
from storops.lib.converter import to_bool
from storops.vnx.enums import VNXLunType, VNXTieringEnum, VNXProvisionEnum, \
    VNXMigrationRate
from storops.vnx.resource import VNXCliResourceList, VNXCliResource
from storops.vnx.resource.migration import VNXMigrationSession
from storops.vnx.resource.snap import VNXSnap, VNXSnapList

__author__ = 'Cedric Zhuang'


class _IsMigratingError(Exception):
    pass


class VNXLunList(VNXCliResourceList):
    def __init__(self, cli=None, lun_type=None, lun_ids=None, pool=None):
        super(VNXLunList, self).__init__(cli)
        self._lun_type = None
        self._lun_ids = None
        self._pool_name = None

        self._set_filter(lun_type, lun_ids, pool)

    def _set_filter(self, lun_type=None, lun_ids=None, pool=None):
        self._lun_type = VNXLunType.parse(lun_type)
        self._lun_ids = lun_ids
        if isinstance(pool, VNXCliResource):
            self._pool_name = pool._get_name()
        else:
            self._pool_name = pool

    def _filter(self, lun):
        if self._lun_ids is not None:
            ret = VNXLun.get_id(lun) in self._lun_ids
        elif self._pool_name is not None:
            ret = lun.pool_name == self._pool_name
        else:
            ret = True
        return ret

    @classmethod
    def get_resource_class(cls):
        return VNXLun

    @property
    @instance_cache
    def _lun_id_map(self):
        return {lun.lun_id: lun for lun in self}

    def _get_raw_resource(self):
        return self._cli.get_lun(lun_type=self._lun_type, poll=self.poll)

    def get(self, _id):
        if isinstance(_id, VNXLun):
            _id = VNXLun.get_id(_id)

        return self._lun_id_map.get(_id)


class VNXLun(VNXCliResource):
    DEFAULT_TIER = VNXTieringEnum.HIGH_AUTO
    DEFAULT_PROVISION = VNXProvisionEnum.THICK

    def __init__(self, lun_id=None, name=None, cli=None):
        super(VNXLun, self).__init__()
        self._cli = cli
        self._lun_id = lun_id
        self._name = name

    def _get_raw_resource(self):
        if self._cli is None:
            raise ValueError('client is not available for this resource.')
        return self._cli.get_lun(name=self._name, lun_id=self._lun_id,
                                 poll=self.poll)

    @property
    def is_snap_mount_point(self):
        return self.primary_lun is not None

    @staticmethod
    def create(cli,
               pool_id=None,
               pool_name=None,
               lun_id=None,
               lun_name=None,
               size_gb=1,
               provision=None,
               tier=None,
               ignore_thresholds=None):
        cls = storops.vnx.resource.block_pool.VNXPool
        pool = cls(pool_id, pool_name, cli)
        return pool.create_lun(lun_name, size_gb, lun_id, provision,
                               tier, ignore_thresholds)

    def create_mount_point(self, _id=None, name=None):
        lun_id = self.get_id(self)
        out = self._cli.create_mount_point(
            primary_lun_id=lun_id,
            mount_point_name=name,
            mount_point_id=_id,
            poll=self.poll)
        ex.raise_if_err(out, default=ex.VNXCreateMpError)
        return VNXLun(lun_id=_id, name=name, cli=self._cli)

    @property
    def tier(self):
        try:
            tier = VNXTieringEnum.get_tier(
                self.initial_tier,
                self.tiering_policy)
        except AttributeError:
            tier = self.DEFAULT_TIER
        return tier

    @tier.setter
    def tier(self, new_tier):
        new_tier = VNXTieringEnum.parse(new_tier)
        out = self._cli.modify_lun(lun_id=self._lun_id,
                                   lun_name=self._name,
                                   new_tier=new_tier,
                                   poll=self.poll)
        msg = 'error change lun tier.'
        ex.raise_if_err(out, msg, default=ex.VNXModifyLunError)

    @property
    def provision(self):
        ret = self.DEFAULT_PROVISION
        try:
            if self.is_thin_lun:
                ret = VNXProvisionEnum.THIN
            if self.is_compressed:
                ret = VNXProvisionEnum.COMPRESSED
            elif self.is_dedup:
                ret = VNXProvisionEnum.DEDUPED
        except AttributeError:
            pass
        return ret

    @property
    def is_dedup(self):
        return to_bool(self.deduplication_state)

    @is_dedup.setter
    def is_dedup(self, value):
        self._update_dedup_state(value)

    @staticmethod
    def get(cli, lun_id=None, name=None, lun_type=None, lun_ids=None,
            poll=True):
        if lun_id is None and name is None:
            ret = VNXLunList(cli=cli, lun_type=lun_type, lun_ids=lun_ids)
        else:
            ret = VNXLun(lun_id, name, cli)
        ret.poll = poll
        return ret

    def create_snap(self, name, allow_rw=None, auto_delete=None,
                    keep_for=None):
        return VNXSnap.create(self._cli, self.get_id(self), name, allow_rw,
                              auto_delete, keep_for)

    def attach_snap(self, snap):
        snap_name = VNXSnap.get_name(snap)
        out = self._cli.attach_snap(snap_name, lun_id=self.get_id(self),
                                    poll=self.poll)
        ex.raise_if_err(out, 'failed to attach snap {}'.format(snap_name),
                        default=ex.VNXAttachSnapError)

    def detach_snap(self):
        lun_id = self.get_id(self)
        out = self._cli.detach_snap(lun_id=lun_id, poll=self.poll)
        msg = 'failed to detach snap for lun {}.'.format(lun_id)
        ex.raise_if_err(out, msg, default=ex.VNXDetachSnapError)

    def get_snap(self, name=None):
        if name is not None:
            ret = VNXSnap.get(self._cli, name)
        else:
            ret = VNXSnapList(self._cli, res=self.get_id(self))
        return ret

    def delete_snap(self, name):
        VNXSnap(name, self._cli).delete()

    def restore_snap(self, name):
        VNXSnap(name, self._cli).restore(self.get_id(self))

    def migrate(self, tgt, rate=VNXMigrationRate.HIGH, on_complete=None,
                on_error=None):
        tgt_id = self.get_id(tgt)
        src_id = self.get_id(self)
        out = self._cli.migrate_lun(src_id, tgt_id, rate, poll=self.poll)
        ex.raise_if_err(out, default=ex.VNXMigrationError)

        if on_complete or on_error:
            ret = daemon(self._wait_for_migration_done, on_complete, on_error)
        else:
            ret = None
        return ret

    @retry(on_error=_IsMigratingError, wait=15, timeout=60 * 60 * 24 * 7)
    def _wait_for_migration_done(self, on_complete=None, on_error=None):
        migration = VNXMigrationSession(source=self, cli=self._cli)
        if migration.is_migrating:
            raise _IsMigratingError()
        elif migration.is_success and on_complete:
            ret = on_complete()
        elif on_error:
            ret = on_error()
        else:
            ret = None
        return ret

    def expand(self, new_size, ignore_thresholds=False):
        out = self._cli.expand_pool_lun(new_size, self.get_id(self),
                                        ignore_thresholds=ignore_thresholds,
                                        poll=self.poll)
        ex.raise_if_err(out, default=ex.VNXLunExtendError)

    def cancel_migrate(self):
        src_id = self.get_id(self)
        out = self._cli.cancel_migrate_lun(src_id, poll=self.poll)
        ex.raise_if_err(out, 'migrate lun {} error.'.format(src_id),
                        default=ex.VNXMigrationError)

    def get_migration_session(self):
        return VNXMigrationSession.get(self._cli, self)

    @staticmethod
    def get_id(lun):
        if isinstance(lun, VNXLun):
            if lun._lun_id is not None:
                lun = lun._lun_id
            else:
                lun = lun.lun_id
        try:
            lun = check_int(lun)
        except ValueError:
            raise ValueError('invalid lun number supplied: {}'
                             .format(lun))
        return lun

    @classmethod
    def get_id_list(cls, *lun_list):
        return list(map(cls.get_id, lun_list))

    def detach_from_sg(self, sg=None):
        if sg is None:
            clz = storops.vnx.resource.sg.VNXStorageGroupList
            obj = clz(cli=self._cli)
        else:
            obj = sg
        obj.detach_alu(self)

    def detach_from_cg(self, cg=None):
        if cg is None:
            clz = storops.vnx.resource.cg.VNXConsistencyGroupList
            obj = clz(cli=self._cli)
        else:
            obj = cg
        obj.delete_member(self)

    def clear_smp(self, force=False):
        smp_list = self.snapshot_mount_points
        if smp_list:
            for smp in self.snapshot_mount_points:
                smp.delete(force=force)

    def delete(self, delete_snapshots=False, force_detach=False,
               detach_from_sg=False, detach_from_cg=False, force=False):
        if force:
            delete_snapshots = True
            force_detach = True
            detach_from_sg = True
            detach_from_cg = True

        while True:
            out = self._do_delete(delete_snapshots, force_detach)
            try:
                ex.raise_if_err(out,
                                'failed to remove lun {}'.format(
                                    self._get_name()),
                                default=ex.VNXDeleteLunError)
                break
            except ex.VNXLunInStorageGroupError:
                if detach_from_sg:
                    self.detach_from_sg()
                else:
                    raise
            except ex.VNXLunInConsistencyGroupError:
                if detach_from_cg:
                    self.detach_from_cg()
                else:
                    raise
            except ex.VNXLunHasSnapMountPointError:
                if force:
                    self.clear_smp(force)
                else:
                    raise

    def _do_delete(self, delete_snapshots, force_detach):
        return self._cli.delete_pool_lun(self._lun_id,
                                         self._get_name(),
                                         delete_snapshots=delete_snapshots,
                                         force_detach=force_detach,
                                         poll=self.poll)

    def rename(self, new_name):
        if new_name is not None and self._name != new_name:
            out = self._cli.modify_lun(lun_id=self._lun_id,
                                       lun_name=self._name,
                                       new_name=new_name,
                                       poll=self.poll)
            ex.raise_if_err(out, 'failed to change lun name.',
                            default=ex.VNXModifyLunError)
            self._name = new_name

    def __setattr__(self, key, value):
        if key == 'name':
            self.rename(value)
            return
        elif key == 'is_compressed':
            if value:
                self.enable_compression()
            else:
                self.disable_compression()
            return
        super(VNXLun, self).__setattr__(key, value)

    def enable_compression(self, rate=None, ignore_thresholds=None):
        lun_id = self.get_id(self)
        out = self._cli.enable_compression(
            lun_id=lun_id, rate=rate, ignore_thresholds=ignore_thresholds,
            poll=self.poll)
        msg = 'failed to enable compression on {}.'.format(lun_id)
        ex.raise_if_err(out, msg, ex.VNXCompressionError)

    def disable_compression(self, ignore_thresholds=None):
        lun_id = self.get_id(self)
        out = self._cli.disable_compression(lun_id, ignore_thresholds,
                                            poll=self.poll)
        ex.raise_if_err(
            out, 'failed to disable compression on {}.'.format(lun_id),
            default=ex.VNXCompressionError)

    def _update_dedup_state(self, tgt_state):
        out = self._cli.modify_lun(lun_id=self._lun_id,
                                   lun_name=self._name,
                                   dedup=tgt_state, poll=self.poll)
        ex.raise_if_err(out, default=ex.VNXDedupError)

    def enable_dedup(self):
        self._update_dedup_state(True)

    def disable_dedup(self):
        self._update_dedup_state(False)

    def create_mirror_view(self, name, use_write_intent_log=True):
        clz = storops.vnx.resource.mirror_view.VNXMirrorView
        return clz.create(self._cli, name, self, use_write_intent_log)

    def get_mirror_view(self, as_src=None, as_tgt=None):
        src_lun = None
        tgt_lun = None
        if as_src is None and as_tgt is None:
            src_lun = self
            tgt_lun = self
        if as_src:
            src_lun = self
        if as_tgt:
            tgt_lun = self

        clz = storops.vnx.resource.mirror_view.VNXMirrorViewList
        return clz(self._cli, src_lun=src_lun, tgt_lun=tgt_lun)



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import inspect
import sys

import six


class ClientException(Exception):
    """The base exception class for all exceptions this library raises."""


class HttpError(ClientException):
    """The base exception class for all HTTP exceptions."""
    http_status = 0
    message = "HTTP Error"

    def __init__(self, message=None, details=None,
                 response=None, request_id=None,
                 url=None, method=None, http_status=None):
        self.http_status = http_status or self.http_status
        self.message = message or self.message
        self.details = details
        self.request_id = request_id
        self.response = response
        self.url = url
        self.method = method
        formatted_string = "%s (HTTP %s)" % (self.message, self.http_status)
        if request_id:
            formatted_string += " (Request-ID: %s)" % request_id
        super(HttpError, self).__init__(formatted_string)


class HttpServerError(HttpError):
    """Server-side HTTP error.

    Exception for cases in which the server is aware that it has
    erred or is incapable of performing the request.
    """
    message = "HTTP Server Error"


class HTTPClientError(HttpError):
    """Client-side HTTP error.

    Exception for cases in which the client seems to have erred.
    """
    message = "HTTP Client Error"


class BadRequest(HTTPClientError):
    """HTTP 400 - Bad Request.

    The request cannot be fulfilled due to bad syntax.
    """
    http_status = 400
    message = "Bad Request"


# _code_map contains all the classes that have http_status attribute.
_code_map = dict(
    (getattr(obj, 'http_status', None), obj)
    for name, obj in six.iteritems(vars(sys.modules[__name__]))
    if inspect.isclass(obj) and getattr(obj, 'http_status', False)
)


def from_response(response, method, url):
    """Returns an instance of :class:`HttpError` or subclass based on response.

    :param response: instance of `requests.Response` class
    :param method: HTTP method used for request
    :param url: URL used for request
    """

    req_id = response.headers.get("x-openstack-request-id")
    # NOTE(hdd) true for older versions of nova and cinder
    if not req_id:
        req_id = response.headers.get("x-compute-request-id")
    kwargs = {
        "http_status": response.status_code,
        "response": response,
        "method": method,
        "url": url,
        "request_id": req_id,
    }
    if "retry-after" in response.headers:
        kwargs["retry_after"] = response.headers["retry-after"]

    content_type = response.headers.get("Content-Type", "")
    if content_type.startswith("application/json"):
        try:
            body = response.json()
        except ValueError:
            pass
        else:
            if isinstance(body, dict):
                error = body.get(list(body)[0])
                if isinstance(error, dict):
                    kwargs["message"] = (error.get("message") or
                                         error.get("faultstring"))
                    kwargs["details"] = (error.get("details") or
                                         six.text_type(body))
    elif content_type.startswith("text/"):
        kwargs["details"] = response.text

    try:
        cls = _code_map[response.status_code]
    except KeyError:
        if 500 <= response.status_code < 600:
            cls = HttpServerError
        elif 400 <= response.status_code < 500:
            cls = HTTPClientError
        else:
            cls = HttpError
    return cls(**kwargs)


class SSHExecutionError(ClientException):
    def __init__(self, stdout=None, stderr=None, exit_code=None, cmd=None,
                 description=None):
        self.exit_code = exit_code
        self.stderr = stderr
        self.stdout = stdout
        self.cmd = cmd
        self.description = description

        if description is None:
            description = "Unexpected error while running command."
        if exit_code is None:
            exit_code = '-'
        message = ('%(description)s\n'
                   'Command: %(cmd)s\n'
                   'Exit code: %(exit_code)s\n'
                   'Stdout: %(stdout)r\n'
                   'Stderr: %(stderr)r') % {'description': description,
                                            'cmd': cmd,
                                            'exit_code': exit_code,
                                            'stdout': stdout,
                                            'stderr': stderr}
        super(SSHExecutionError, self).__init__(message)


class SFtpExecutionError(ClientException):
    message = "[EMC] Failed to execute sftp operation. %(err)s."



# coding=utf-8
# Copyright (c) 2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import unicode_literals

import logging

from storops.unity.enums import ScheduleTypeEnum, DayOfWeekEnum, \
    FilesystemSnapAccessTypeEnum
from storops.unity.resource import UnityResource, UnityResourceList

__author__ = 'Ryan Liang'

log = logging.getLogger(__name__)


class UnitySnapScheduleRule(UnityResource):
    @staticmethod
    def to_embedded(schedule_type, minute=None, hours=None, days_of_week=None,
                    days_of_month=None, interval=None, is_auto_delete=None,
                    retention_time=None, access_type=None):
        """
        Constructs a `SnapScheduleRule` embedded instance.

        The rule created depends on the value of `schedule_type`. Its value
        could be:
        0. `N_HOURS_AT_MM`, snap every `interval` hours, at `minute`
            past the hour. Supported parameters: `interval` (required),
            `minute` (optional, default 0).
        1. `DAY_AT_HHMM`, specify `hours` as a list of `hour[,...]` to snap one
            or more times each day at `minute` past the hour. Supported
            parameters: `hours` (at least one required), `minute` (optional).
        2. `N_DAYS_AT_HHMM`, snap every `interval` days at the time
            `hours`:`minute`. Supported Parameters: `interval` (required),
            `hours` (optional, exactly one), `minute` (optional).
        3. `SELDAYS_AT_HHMM`, snap on the selected `days_of_week`, at the time
            `hours`:`minute`. Supported parameters: `days_of_week` (at least
            one required), `hours` (optional, default 0), `minute` (optional,
            default 0).
        4. `NTH_DAYOFMONTH_AT_HHMM`, snap on the selected `days_of_month`, at
            the time `hours`:`minute`. Supported parameters: `days_of_month`
            (at least one required), `hours` (optional, default 0), `minute`
            (optional, default 0).
        5. `UNSUPPORTED`.

        :param schedule_type: Type of snapshot schedule rule. Value is
            ScheduleTypeEnum.
        :param minute: Minute frequency for the snapshot schedule rule. Value
            should be [0, 59].
        :param hours: Hourly frequency for the snapshot schedule rule. Each
            value should be [0, 23].
        :param days_of_week: Days of the week for which the snapshot schedule
            rule applies. Each value is DayOfWeekEnum.
        :param days_of_month: Days of the month for which the snapshot schedule
            rule applies. Each value should be [1, 31].
        :param interval: Number of days or hours between snaps, depending on
            the rule type.
        :param is_auto_delete: Indicates whether the system can automatically
            delete the snapshot based on pool automatic-deletion thresholds.
            Values are:
                true - System can delete the snapshot based on pool
                    automatic-deletion thresholds.
                false - System cannot delete the snapshot based on pool
                    automatic-deletion thresholds.
        :param retention_time: (Applies when the value of the is_auto_delete
            attribute is false.) Period of time in seconds for which to keep
            the snapshot.
        :param access_type: For a file system or VMware NFS datastore snapshot
            schedule, indicates whether the snapshot created by the schedule
            has checkpoint or protocol type access. Value is
            FilesystemSnapAccessTypeEnum.
        :return: constructed embedded instance in a dict.
        """
        ScheduleTypeEnum.verify(schedule_type, allow_none=False)
        if minute is not None and (minute < 0 or minute > 59):
            raise ValueError(
                'minute in UnitySnapScheduleRule should be [0, 59].')
        if hours is not None:
            if not isinstance(hours, list) or not hours:
                raise ValueError(
                    'hours in UnitySnapScheduleRule should be a list '
                    'and not an empty list.')
            if any([(hour < 0 or hour > 23) for hour in hours]):
                raise ValueError(
                    'each value of hours in UnitySnapScheduleRule should be '
                    '[0, 23].')
        if days_of_week is not None:
            if not isinstance(days_of_week, list) or not days_of_week:
                raise ValueError(
                    'days_of_week in UnitySnapScheduleRule should be a list '
                    'and not an empty list.')
            for day_of_week in days_of_week:
                DayOfWeekEnum.verify(day_of_week, allow_none=False)

        if days_of_month is not None:
            if not isinstance(days_of_month, list) or not days_of_month:
                raise ValueError(
                    'days_of_month in UnitySnapScheduleRule should be a list '
                    'and not an empty list.')
            if any([(day < 1 or day > 31) for day in days_of_month]):
                raise ValueError(
                    'each value of days_of_month in UnitySnapScheduleRule '
                    'should be [1, 31].')

        FilesystemSnapAccessTypeEnum.verify(access_type)

        return {
            'type': schedule_type,
            'minute': minute,
            'hours': hours,
            'daysOfWeek': days_of_week,
            'daysOfMonth': days_of_month,
            'interval': interval,
            'isAutoDelete': is_auto_delete,
            'retentionTime': retention_time,
            'accessType': access_type,
        }

    @staticmethod
    def every_n_hours(hour_interval, minute=0, is_auto_delete=None,
                      retention_time=None, access_type=None):
        """
        Helper function to construct `N_HOURS_AT_MM` type rule.

        In this rule the snapshot will be taken every `hour_interval` hours,
        at `minute` past the hour.

        :param hour_interval: the interval in hour the snapshot will be taken.
        :param minute: minutes past the hour when the snapshot will be taken.
            Value should be [0, 59], and is 0 by default.
        :param is_auto_delete: Indicates whether the system can automatically
            delete the snapshot based on pool automatic-deletion thresholds.
            Values are:
                true - System can delete the snapshot based on pool
                    automatic-deletion thresholds.
                false - System cannot delete the snapshot based on pool
                    automatic-deletion thresholds.
        :param retention_time: (Applies when the value of the is_auto_delete
            attribute is false.) Period of time in seconds for which to keep
            the snapshot.
        :param access_type: For a file system or VMware NFS datastore snapshot
            schedule, indicates whether the snapshot created by the schedule
            has checkpoint or protocol type access. Value is
            FilesystemSnapAccessTypeEnum.
        :return: constructed embedded instance in a dict.
        """
        return UnitySnapScheduleRule.to_embedded(
            ScheduleTypeEnum.N_HOURS_AT_MM,
            interval=hour_interval,
            minute=minute,
            is_auto_delete=is_auto_delete,
            retention_time=retention_time,
            access_type=access_type,
        )

    @staticmethod
    def every_day(hours, minute=0, is_auto_delete=None, retention_time=None,
                  access_type=None):
        """
        Helper function to construct `DAY_AT_HHMM` type rule.

        In this rule the snapshot will be taken at hours specified by a list of
        `hour[,...]`, and at `minute` past the hour.

        :param hours: a list of `hour[,...]` at which hour the snapshot will be
            taken. Cannot be empty list and each value should be [0, 23].
        :param minute: minutes past the hour when the snapshot will be taken.
            Value should be [0, 59], and is 0 by default.
        :param is_auto_delete: Indicates whether the system can automatically
            delete the snapshot based on pool automatic-deletion thresholds.
            Values are:
                true - System can delete the snapshot based on pool
                    automatic-deletion thresholds.
                false - System cannot delete the snapshot based on pool
                    automatic-deletion thresholds.
        :param retention_time: (Applies when the value of the is_auto_delete
            attribute is false.) Period of time in seconds for which to keep
            the snapshot.
        :param access_type: For a file system or VMware NFS datastore snapshot
            schedule, indicates whether the snapshot created by the schedule
            has checkpoint or protocol type access. Value is
            FilesystemSnapAccessTypeEnum.
        :return: constructed embedded instance in a dict.
        """
        if not isinstance(hours, list):
            raise ValueError(
                'hours of UnitySnapScheduleRule every_day rule should be '
                'a list.')
        if not hours:
            raise ValueError(
                'hours of UnitySnapScheduleRule every_day rule cannot be '
                'empty list.')
        return UnitySnapScheduleRule.to_embedded(
            ScheduleTypeEnum.DAY_AT_HHMM,
            interval=1,  # Every day
            hours=hours,
            minute=minute,
            is_auto_delete=is_auto_delete,
            retention_time=retention_time,
            access_type=access_type,
        )

    @staticmethod
    def every_n_days(day_interval, hour=0, minute=0, is_auto_delete=None,
                     retention_time=None, access_type=None):
        """
        Helper function to construct `N_DAYS_AT_HHMM` type rule.

        In this rule the snapshot will be taken every `day_interval` days at
        the time `hour`:`minute`.

        :param day_interval: the interval in day the snapshot will be taken.
        :param hour: the hour when the snapshot will be taken. Value should be
            [0, 23], and is 0 by default.
        :param minute: minutes past the hour when the snapshot will be taken.
            Value should be [0, 59], and is 0 by default.
        :param is_auto_delete: Indicates whether the system can automatically
            delete the snapshot based on pool automatic-deletion thresholds.
            Values are:
                true - System can delete the snapshot based on pool
                    automatic-deletion thresholds.
                false - System cannot delete the snapshot based on pool
                    automatic-deletion thresholds.
        :param retention_time: (Applies when the value of the is_auto_delete
            attribute is false.) Period of time in seconds for which to keep
            the snapshot.
        :param access_type: For a file system or VMware NFS datastore snapshot
            schedule, indicates whether the snapshot created by the schedule
            has checkpoint or protocol type access. Value is
            FilesystemSnapAccessTypeEnum.
        :return: constructed embedded instance in a dict.
        """
        return UnitySnapScheduleRule.to_embedded(
            ScheduleTypeEnum.N_DAYS_AT_HHMM,
            interval=day_interval,
            hours=[hour],
            minute=minute,
            is_auto_delete=is_auto_delete,
            retention_time=retention_time,
            access_type=access_type,
        )

    @staticmethod
    def every_week(days_of_week, hour=0, minute=0, is_auto_delete=None,
                   retention_time=None, access_type=None):
        """
        Helper function to construct `SELDAYS_AT_HHMM` type rule.

        In this rule the snapshot will be taken on the selected `days_of_week`,
        at the time `hour`:`minute`.

        :param days_of_week: days in every week the snapshot will be taken.
            Cannot be empty list and each value is DayOfWeekEnum.
        :param hour: the hour when the snapshot will be taken. Value should be
            [0, 23], and is 0 by default.
        :param minute: minutes past the hour when the snapshot will be taken.
            Value should be [0, 59], and is 0 by default.
        :param is_auto_delete: Indicates whether the system can automatically
            delete the snapshot based on pool automatic-deletion thresholds.
            Values are:
                true - System can delete the snapshot based on pool
                    automatic-deletion thresholds.
                false - System cannot delete the snapshot based on pool
                    automatic-deletion thresholds.
        :param retention_time: (Applies when the value of the is_auto_delete
            attribute is false.) Period of time in seconds for which to keep
            the snapshot.
        :param access_type: For a file system or VMware NFS datastore snapshot
            schedule, indicates whether the snapshot created by the schedule
            has checkpoint or protocol type access. Value is
            FilesystemSnapAccessTypeEnum.
        :return: constructed embedded instance in a dict.
        """
        if not isinstance(days_of_week, list):
            raise ValueError(
                'days_of_week of UnitySnapScheduleRule every_week rule should '
                'be a list.')
        if not days_of_week:
            raise ValueError(
                'days_of_week of UnitySnapScheduleRule every_week rule cannot '
                'be empty list.')
        return UnitySnapScheduleRule.to_embedded(
            ScheduleTypeEnum.SELDAYS_AT_HHMM,
            days_of_week=days_of_week,
            hours=[hour],
            minute=minute,
            is_auto_delete=is_auto_delete,
            retention_time=retention_time,
            access_type=access_type,
        )

    @staticmethod
    def every_month(days_of_month, hour=0, minute=0, is_auto_delete=None,
                    retention_time=None, access_type=None):
        """
        Helper function to construct `NTH_DAYOFMONTH_AT_HHMM` type rule.

        In this rule the snapshot will be taken on the selected
        `days_of_month`, at the time `hour`:`minute`.

        :param days_of_month: days in every month the snapshot will be taken.
            Cannot be empty list.
        :param hour: the hour when the snapshot will be taken. Value should be
            [0, 23], and is 0 by default.
        :param minute: minutes past the hour when the snapshot will be taken.
            Value should be [0, 59], and is 0 by default.
        :param is_auto_delete: Indicates whether the system can automatically
            delete the snapshot based on pool automatic-deletion thresholds.
            Values are:
                true - System can delete the snapshot based on pool
                    automatic-deletion thresholds.
                false - System cannot delete the snapshot based on pool
                    automatic-deletion thresholds.
        :param retention_time: (Applies when the value of the is_auto_delete
            attribute is false.) Period of time in seconds for which to keep
            the snapshot.
        :param access_type: For a file system or VMware NFS datastore snapshot
            schedule, indicates whether the snapshot created by the schedule
            has checkpoint or protocol type access. Value is
            FilesystemSnapAccessTypeEnum.
        :return: constructed embedded instance in a dict.
        """
        if not isinstance(days_of_month, list):
            raise ValueError(
                'days_of_month of UnitySnapScheduleRule every_month rule '
                'should be a list.')
        if not days_of_month:
            raise ValueError(
                'days_of_month of UnitySnapScheduleRule every_month rule '
                'cannot be empty list.')
        return UnitySnapScheduleRule.to_embedded(
            ScheduleTypeEnum.NTH_DAYOFMONTH_AT_HHMM,
            days_of_month=days_of_month,
            hours=[hour],
            minute=minute,
            is_auto_delete=is_auto_delete,
            retention_time=retention_time,
            access_type=access_type,
        )


class UnitySnapScheduleRuleList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnitySnapScheduleRule


class UnitySnapSchedule(UnityResource):
    @classmethod
    def create(cls, cli, name, rules,
               is_sync_replicated=None, skip_sync_to_remote_system=None):
        """
        :param cli:
        :param name: Name of new schedule.
        :param rules: Rules that apply to the snapshot schedule, as defined by
            the `UnitySnapScheduleRule` resource type. Each value is a
            `UnitySnapScheduleRule` instance.
        :param is_sync_replicated: Indicates that all operations on the
            snapshot schedule will be synchronously replicated to the peer
            system.
        :param skip_sync_to_remote_system: For internal system use only. Skip
            the invocation of the snapshot schedule creation method with the
            same parameters on the remote system. If the method is invoked by
            the user, this attribute is not required. The value should be true
            when the operation is executed from a remote system, to prevent
            from syncing it back to the original storage.
        :return: the created `UnitySnapSchedule` instance.
        """

        resp = cli.post(cls().resource_class,
                        name=name,
                        rules=rules,
                        isSyncReplicated=is_sync_replicated,
                        skipSyncToRemoteSystem=skip_sync_to_remote_system)
        resp.raise_if_err()
        return cls(_id=resp.resource_id, cli=cli)

    def modify(self, add_rules=None, remove_rule_ids=None,
               skip_sync_to_remote_system=None):
        """
        :param add_rules: Rules to add to the snapshot schedule, as defined by
            the UnitySnapScheduleRule instance. Pass the desired values for the
             additional rules instead of the unique identifiers of existing
             rules. Each value is a `UnitySnapScheduleRule` instance.
        :param remove_rule_ids: Unique identifiers of the rules to remove from
            the snapshot schedule.
        :param skip_sync_to_remote_system: For internal system use only. Skip
            the invocation of the snapshot schedule modification method with
            the same parameters on the remote system. If the method is invoked
            by the user, this attribute is not required. The value should be
            true when the operation is executed from a remote system, to
            prevent from syncing it back to the original storage.
        :return:
        """
        req_body = self._cli.make_body(
            addRules=add_rules, removeRuleIds=remove_rule_ids,
            skipSyncToRemoteSystem=skip_sync_to_remote_system)
        resp = self.action('modify', **req_body)
        resp.raise_if_err()
        return resp


class UnitySnapScheduleList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnitySnapSchedule



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging
import re

from storops.lib.common import Enum, cache

log = logging.getLogger(__name__)


class VNXEnum(Enum):
    pass


class VNXSPEnum(VNXEnum):
    SP_A = 'SP A'
    SP_B = 'SP B'
    CONTROL_STATION = 'Celerra'

    @classmethod
    def get_int_index(cls):
        return None, cls.SP_A, cls.SP_B, cls.CONTROL_STATION

    @classmethod
    @cache
    def get_to_delete(cls):
        return re.compile('[_. ]')

    @classmethod
    def is_sp(cls, name):
        name = cls.parse(name)
        return name in (cls.SP_A, cls.SP_B)

    @classmethod
    def _normalize(cls, value):
        ret = re.sub(cls.get_to_delete(), '', value)
        if ret is None:
            pass
        elif ret.endswith('a') and not ret.endswith('rra'):
            ret = 'spa'
        elif ret.endswith('b'):
            ret = 'spb'
        elif ret == 'cs':
            ret = 'celerra'
        return ret

    @classmethod
    def from_str(cls, value):
        value = value.lower()
        value = cls._normalize(value)
        ret = None
        for item in cls.get_all():
            if cls._normalize(item.value.lower()) in value:
                ret = item
                break
        else:
            log.info('cannot parse "{}" to a vnx sp.'.format(value))
        return ret

    @classmethod
    def get_sp_index(cls, value):
        value = cls.parse(value)
        if value is None:
            raise ValueError('"{}" is not a valid sp name.'.format(value))
        return value.index

    @property
    def index(self):
        return self.value.lower()[-1]

    @property
    def display_name(self):
        return self.index.upper()


class VNXProvisionEnum(VNXEnum):
    # value of spec "provisioning:type"
    THIN = 'thin'
    THICK = 'thick'
    COMPRESSED = 'compressed'
    DEDUPED = 'deduplicated'

    @classmethod
    def get_option_map(cls):
        return {
            cls.THIN: ['-type', 'Thin'],
            cls.THICK: ['-type', 'NonThin'],
            cls.COMPRESSED: ['-type', 'Thin'],
            cls.DEDUPED: ['-type', 'Thin', '-deduplication', 'on']}


class VNXCompressionRate(VNXEnum):
    LOW = 'low'
    MEDIUM = 'medium'
    HIGH = 'high'


class VNXTieringPreference(VNXEnum):
    INVALID = 0
    NONE = 1
    LOWEST_AVAILABLE = 2
    HIGHEST_AVAILABLE = 3


class VNXRelocationPolicy(VNXEnum):
    INVALID = 0
    NONE = 1
    TIER_PREFERENCE = 2
    OPTIMAL = 3


class VNXMigrationRate(VNXEnum):
    LOW = 'low'
    MEDIUM = 'medium'
    HIGH = 'high'
    ASAP = 'asap'


class VNXTieringEnum(VNXEnum):
    NONE = 'none'
    HIGH_AUTO = 'starthighthenauto'
    AUTO = 'auto'
    HIGH = 'highestavailable'
    LOW = 'lowestavailable'
    NO_MOVE = 'nomovement'

    @classmethod
    def get_option_map(cls):
        return {
            cls.NONE: [],
            cls.HIGH_AUTO: [
                '-initialTier', 'highestAvailable',
                '-tieringPolicy', 'autoTier'],
            cls.AUTO: [
                '-initialTier', 'optimizePool',
                '-tieringPolicy', 'autoTier'],
            cls.HIGH: [
                '-initialTier', 'highestAvailable',
                '-tieringPolicy', 'highestAvailable'],
            cls.LOW: [
                '-initialTier', 'lowestAvailable',
                '-tieringPolicy', 'lowestAvailable'],
            cls.NO_MOVE: [
                '-initialTier', 'optimizePool',
                '-tieringPolicy', 'noMovement']
        }

    @classmethod
    def get_tier(cls, initial, policy):
        ret = None
        for k, v in cls.get_option_map().items():
            if len(v) >= 4:
                v_initial, v_policy = v[1], v[3]
                if (cls.match_option(initial, v_initial) and
                        cls.match_option(policy, v_policy)):
                    ret = k
                    break
                elif cls.match_option(policy, 'noMovement'):
                    """no movement could have different initial tier"""
                    ret = cls.NO_MOVE
                    break
        if ret is None:
            raise ValueError('Initial tier: {}, policy: {} is not valid.'
                             .format(initial, policy))
        return ret

    @staticmethod
    def match_option(output, option):
        return output.replace(' ', '').lower() == option.lower()


class VNXPortType(VNXEnum):
    FC = 'FC'
    ISCSI = 'iSCSI'
    FCOE = 'FCoE'
    SAS = 'SAS'
    ETHERNET = 'ethernet'
    OTHER = 'other'

    @classmethod
    def get_int_index(cls):
        return None, cls.FC, cls.ISCSI, cls.FCOE, cls.SAS

    @classmethod
    def from_str(cls, value):
        if '.' in value:
            value = 'iSCSI'
        elif ':' in value:
            value = 'FC'

        ret = None
        if value is not None:
            for item in cls.get_all():
                if item.is_equal(value):
                    ret = item
                    break
            else:
                cls._raise_invalid_value(value)
        return ret


class VNXSnapType(VNXEnum):
    LUN = 1
    CG = 2


class VNXMirrorViewRecoveryPolicy(VNXEnum):
    MANUAL = 'manual'
    AUTO = 'automatic'

    @classmethod
    def get_option_map(cls):
        return {
            cls.MANUAL: ['-recoverypolicy', 'manual'],
            cls.AUTO: ['-recoverypolicy', 'auto']
        }


VNXMirrorGroupRecoveryPolicy = VNXMirrorViewRecoveryPolicy


class VNXMirrorViewSyncRate(VNXEnum):
    HIGH = 'high'
    MEDIUM = 'medium'
    LOW = 'low'


class VNXLunType(VNXEnum):
    THIN = 'Thin'
    NON_THIN = 'NonThin'
    SNAP = 'Snap'
    SNAP_MOUNT_POINT = 'Snap'
    COMPRESSED = 'Compressed'
    NON_COMPRESSED = 'NonCompressed'
    DEDUPED = 'Deduped'
    NON_DEDUPED = 'NonDeduped'


class VNXRaidType(VNXEnum):
    RAID0 = 'r0'
    RAID1 = 'r1'
    RAID2 = 'r2'
    RAID3 = 'r3'
    RAID4 = 'r4'
    RAID5 = 'r5'
    RAID6 = 'r6'
    RAID10 = 'r1_0'
    HOTSPARE = 'hot_spare'
    UNBOUND = 'Unbound'


class VNXPoolRaidType(VNXEnum):
    RAID0 = 'r_0'
    RAID1 = 'r_1'
    RAID5 = 'r_5'
    RAID6 = 'r_6'
    RAID10 = 'r_10'
    RAID_MIXED = 'Mixed'

    @property
    def min_disk_requirement(self):
        clz = VNXPoolRaidType
        if self is clz.RAID0:
            ret = 2
        elif self is clz.RAID1:
            ret = 2
        elif self is clz.RAID5:
            ret = 3
        elif self is clz.RAID6:
            ret = 4
        elif self is clz.RAID10:
            ret = 4
        else:
            raise ValueError('invalid VNXPoolRaidType supplied.')
        return ret


class VNXCtrlMethod(object):
    NO_CTRL = 'noctrl'
    LIMIT_CTRL = 'limit'
    CRUISE_CTRL = 'cruise'
    FIXED_CTRL = 'fixed'

    def __init__(self, method, metric=None, value=None, tolerance=None):
        self.method = method
        self.metric = metric
        self.value = value
        self.tolerance = tolerance

    def get_option(self):
        if self.method == VNXCtrlMethod.NO_CTRL:
            option = ['-noctrl']
        elif self.method == VNXCtrlMethod.LIMIT_CTRL:
            option = ['-ctrlmethod', self.method, '-gmetric', self.metric,
                      '-gval', self.value]
        elif self.method == VNXCtrlMethod.CRUISE_CTRL:
            option = ['-ctrlmethod', self.method, '-gmetric', self.metric,
                      '-gval', self.value, '-gtol', self.tolerance]
        elif self.method == VNXCtrlMethod.FIXED_CTRL:
            option = ['-ctrlmethod', self.method, '-gval', self.value]
        elif not self.method:
            option = []
        else:
            raise ValueError(
                'Invalid control method specified. %s' % self.method)
        return option

    @staticmethod
    def parse_cmd(ctrlmethod):
        if not ctrlmethod:
            ctrlmethod = VNXCtrlMethod(method=None)
        elif ctrlmethod == VNXCtrlMethod.NO_CTRL:
            ctrlmethod = VNXCtrlMethod(method=ctrlmethod)
        cmd = ctrlmethod.get_option()
        return cmd


class VNXAccessLevel(VNXEnum):
    RW = 'rw'
    RO = 'ro'
    ACCESS = 'access'
    ROOT = 'root'


class VNXShareType(VNXEnum):
    NFS = 'nfs'
    CIFS = 'cifs'


class VNXUserScopeEnum(VNXEnum):
    GLOBAL = 'global'
    LOCAL = 'local'


class VNXUserRoleEnum(VNXEnum):
    ADMIN = 'administrator'
    STORAGE_ADMIN = 'storageadmin'
    OPERATOR = 'operator'
    SECURITY_ADMIN = 'securityadministrator'
    DATA_PROTECTION = 'dataprotection'
    LOCAL_DATA_PROTECTION = 'localdataprotection'
    DATA_RECOVERY = 'datarecovery'
    SAN_ADMIN = 'sanadmin'
    NETWORK_ADMIN = 'networkadmin'
    NAS_ADMIN = 'nasadmin'
    VM_ADMIN = 'vmadmin'


class VNXMirrorImageState(VNXEnum):
    SYNCHRONIZED = 'Synchronized'
    OUT_OF_SYNC = 'Out-of-Sync'
    SYNCHRONIZING = 'Synchronizing'
    CONSISTENT = 'Consistent'
    SCRAMBLED = 'Scrambled'
    INCOMPLETE = 'Incomplete'
    LOCAL_ONLY = 'Local Only'
    EMPTY = 'Empty'



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from storops.exception import get_rest_exception
from storops.unity.resource import health
from storops.unity.resource import job
from storops.lib.common import instance_cache

__author__ = 'Cedric Zhuang'


class RestResponse(object):
    def __init__(self, inputs, cli=None):
        if isinstance(inputs, (tuple, list)):
            if len(inputs) == 2:
                self.response = inputs[0]
                self.body = inputs[1]
            elif len(inputs) == 1:
                self.response = None
                self.body = inputs[0]
            else:
                raise ValueError('not valid input: {}'.format(inputs))
        else:
            self.response = None
            self.body = inputs
        if not self.body:
            self.body = {}
        self._cli = cli

    @property
    def contents(self):
        if 'entries' in self.body:
            ret = [entry.get('content', {})
                   for entry in self.body['entries']]
        elif 'content' in self.body:
            ret = [self.body.get('content')]
        else:
            ret = []
        return ret

    @property
    def entries(self):
        return self.body.get('entries', [])

    @property
    def first_content(self):
        contents = self.contents
        if contents:
            ret = contents[0]
        else:
            ret = {}
        return ret

    @property
    def resource_id(self):
        ret = self.first_content.get('id')
        if ret is None:
            sr = self.first_content.get('storageResource')
            if sr is not None:
                ret = sr.get('id')
        return ret

    def has_error(self):
        return self.error is not None

    def is_ok(self):
        return not self.has_error()

    @property
    def has_next_page(self):
        return self.next_page is not None

    @property
    def next_page(self):
        return self._get_page_number('next')

    @property
    def current_page(self):
        return self._get_page_number('self')

    def _get_page_number(self, which_page):
        links = self.body.get('links')
        ret = None
        if links:
            page_link = list(filter(lambda l: l.get('rel') == which_page,
                                    links))
            if page_link:
                href = page_link[0].get('href')
                if href:
                    items = href.split('=')
                    if len(items) > 1:
                        ret = int(items[1])
        return ret

    @property
    @instance_cache
    def error(self):
        clz = health.UnityError
        err = self.body.get('error')
        if err is not None:
            ret = clz().update(err)
        else:
            ret = None
        return ret

    @property
    def error_code(self):
        if self.has_error():
            ret = self.error.error_code
        else:
            ret = None
        return ret

    def raise_if_err(self):
        if self.has_error():
            ex_clz = get_rest_exception(self.error_code)
            raise ex_clz(self.error)

    @property
    def job(self):
        if not hasattr(self, '_job'):
            self._job = job.UnityJob(cli=self._cli)
            self._job.update(self.body)
        return self._job

    @job.setter
    def job(self, value):
        self._job = value


RESP_OK = RestResponse({})



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from storops.lib.common import instance_cache, clear_instance_cache
from storops.vnx.resource import VNXCliResource, VNXCliResourceList
from storops.vnx.resource.sg import VNXStorageGroupList

__author__ = 'Cedric Zhuang'


class VNXHost(VNXCliResource):
    def __init__(self, name=None, cli=None):
        super(VNXHost, self).__init__()
        self._cli = cli
        self._name = name

        self.name = self._name
        self.connections = None
        self.storage_group = None
        self._existed = False

    def property_names(self):
        return ['name', 'connections', 'lun_list', 'alu_hlu_map']

    def get_index(self):
        return 'name'

    @property
    @instance_cache
    def lun_list(self):
        return self.storage_group.lun_list

    @property
    @instance_cache
    def alu_hlu_map(self):
        return self.storage_group.get_alu_hlu_map()

    @property
    def alu_ids(self):
        if self.alu_hlu_map is not None:
            ret = self.alu_hlu_map.keys()
        else:
            ret = tuple()
        return ret

    @property
    def hlu_ids(self):
        if self.alu_hlu_map is not None:
            ret = self.alu_hlu_map.values()
        else:
            ret = tuple()
        return ret

    @clear_instance_cache
    def update(self, data=None):
        host_list = VNXHostList(cli=self._cli)
        for host in host_list:
            if host.name == self._name:
                self._existed = True
                self.name = host.name
                self.connections = host.connections
                self.storage_group = host.storage_group
        return self

    @property
    def existed(self):
        return self._existed

    @classmethod
    def get(cls, cli, name=None):
        if name is None:
            ret = VNXHostList(cli)
        else:
            ret = VNXHost(cli=cli, name=name).update()
        return ret


class VNXHostList(VNXCliResourceList):
    def __init__(self, cli=None, names=None):
        super(VNXHostList, self).__init__()
        self._cli = cli
        self._names = names

    @classmethod
    def get_resource_class(cls):
        return VNXHost

    def update(self, data=None):
        sg_list = VNXStorageGroupList(cli=self._cli, engineering=True)
        hosts = []
        host_names = []
        for sg in sg_list:
            if sg.hba_sp_pairs:
                for pair in sg.hba_sp_pairs:
                    name = pair.host_name
                    if self._names is not None and name not in self._names:
                        continue

                    if name in host_names:
                        host = hosts[host_names.index(name)]
                    else:
                        host = VNXHost(name=name, cli=self._cli)
                        host.connections = []
                        host.storage_group = sg
                        hosts.append(host)
                        host_names.append(name)

                    host.connections.append(pair)
        self._list = hosts
        return self



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, equal_to
from storops.vnx.resource.capacity import VNXCapacity
from storops_test.vnx.cli_mock import patch_cli, t_cli

__author__ = 'Tina Tang'


class VNXCapacityTest(TestCase):
    @patch_cli
    def test_total(self):
        capacity = VNXCapacity.get(t_cli())
        assert_that(capacity.total, equal_to(178269.891))

    @patch_cli
    def test_free_raw_disk(self):
        capacity = VNXCapacity.get(t_cli())
        assert_that(capacity.free_raw_disk, equal_to(42170.744))

    @patch_cli
    def test_free_storage_pool(self):
        capacity = VNXCapacity.get(t_cli())
        assert_that(capacity.free_storage_pool, equal_to(11958.398))

    @patch_cli
    def test_used(self):
        capacity = VNXCapacity.get(t_cli())
        assert_that(capacity.used, equal_to(124140.749))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import unittest

import hamcrest as hc

import storops
from storops.unity.resource import migration_object, remote_system
from storops_test.unity import rest_mock as mock

__author__ = 'Yong Huang'


class UnityMoveSessionTest(unittest.TestCase):
    @mock.patch_rest
    def test_get_properties(self):
        instance_id = 'root/emc:EMC_UIS_MigrationResourceLeaf\"%InstanceID=' \
                      'mo_17628%Name=APM00152904558%SystemName=VNX\"'
        wwn = '60:06:01:60:12:90:38:00:F7:99:07:16:BE:DF:E7:11'
        mo = migration_object.UnityMigrationObject.get(cli=mock.t_rest(),
                                                       _id='mo_17628')
        hc.assert_that(mo.id, hc.equal_to('mo_17628'))
        hc.assert_that(mo.type, hc.equal_to(storops.RemoteObjectTypeEnum.LUN))
        hc.assert_that(mo.instance_id, hc.equal_to(instance_id))
        hc.assert_that(mo.remote_id, hc.equal_to('541'))
        hc.assert_that(mo.name, hc.equal_to('lun-liangr'))
        hc.assert_that(mo.lun_info.is_thin, hc.equal_to(False))
        hc.assert_that(mo.lun_info.size, hc.equal_to(10737418240))
        hc.assert_that(mo.lun_info.wwn, hc.equal_to(wwn))
        hc.assert_that(mo.lun_info.storage_pool_name,
                       hc.equal_to('DAILY_1'))
        hc.assert_that(mo.lun_info.default_sp, hc.equal_to('1'))
        hc.assert_that(mo.remote_system.id, hc.equal_to('RS_65557'))

    @mock.patch_rest
    def test_get_collection(self):
        mo_list = migration_object.UnityMigrationObjectList.get(
            mock.t_rest())
        hc.assert_that(len(mo_list), hc.equal_to(3))
        hc.assert_that([mo.id for mo in mo_list],
                       hc.equal_to(['mo_8182', 'mo_17628', 'mo_17593']))

    @mock.patch_rest
    def test_discover(self):
        remote_sys = remote_system.UnityRemoteSystem(_id='RS_65557',
                                                     cli=mock.t_rest())
        resp = migration_object.UnityMigrationObject.discover(mock.t_rest(),
                                                              remote_sys)
        hc.assert_that(resp.is_ok(), hc.equal_to(True))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging

import storops.unity.resource.cifs_share
import storops.unity.resource.nfs_share
import storops.unity.resource.storage_resource
from storops.exception import UnityCGMemberActionNotSupportError, \
    UnityThinCloneNotAllowedError, UnityDeleteAttachedSnapError
from storops.lib.common import instance_cache
from storops.lib.thinclone_helper import TCHelper
from storops.lib.version import version
from storops.unity import enums
from storops.unity.enums import FilesystemSnapAccessTypeEnum, SnapStateEnum, \
    SnapAccessLevelEnum
from storops.unity.resource import UnityResource, UnityResourceList

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


class UnitySnap(UnityResource):
    @classmethod
    def create(cls, cli, storage_resource, name=None,
               description=None, is_auto_delete=None,
               retention_duration=None, is_read_only=None,
               fs_access_type=None):
        FilesystemSnapAccessTypeEnum.verify(fs_access_type)
        sr_clz = storops.unity.resource.storage_resource.UnityStorageResource
        sr = sr_clz.get(cli, storage_resource)

        resp = cli.post(cls().resource_class,
                        storageResource=sr,
                        name=name,
                        description=description,
                        isAutoDelete=is_auto_delete,
                        retentionDuration=retention_duration,
                        isReadOnly=is_read_only,
                        filesystemAccessType=fs_access_type)
        resp.raise_if_err()
        return cls(_id=resp.resource_id, cli=cli)

    def create_cifs_share(self, name, path=None, is_read_only=None):
        clz = storops.unity.resource.cifs_share.UnityCifsShare
        return clz.create_from_snap(
            self._cli, snap=self, name=name, path=path,
            is_read_only=is_read_only)

    def create_nfs_share(self, name, path=None, is_read_only=None,
                         default_access=None):
        clz = storops.unity.resource.nfs_share.UnityNfsShare
        return clz.create_from_snap(
            self._cli, snap=self, name=name, path=path,
            is_read_only=is_read_only, default_access=default_access)

    def create_snap(self, name):
        return self.copy(copy_name=name)

    def copy(self, copy_name=None):
        resp = self.action('copy', copyName=copy_name)
        resp.raise_if_err()
        try:
            snap_copy_id = resp.contents[0]['copies'][0]['id']
            ret = self.get(self._cli, snap_copy_id)
        except (IndexError, KeyError, TypeError):
            log.exception('failed to get snap id from resp: {}'.format(resp))
            raise
        return ret

    @property
    @instance_cache
    def filesystem(self):
        sr = self.storage_resource
        if sr:
            ret = sr.filesystem
        else:
            ret = None
        return ret

    @property
    def existed(self):
        return (super(UnitySnap, self).existed and
                self.state != SnapStateEnum.DESTROYING)

    def _get_group_snap_id(self):
        if self.is_member_snap():
            # Member snap of cg snap isn't allowed to attach directly,
            # use its group snap to attach.
            cg_snap_id = self.snap_group.get_id()
            log.info('Snap %s is a member snap of cg snap. Use cg '
                     'snap %s to attach/detach.', self.get_id(), cg_snap_id)
            return cg_snap_id
        return self.get_id()

    @version('>=4.1')
    def attach_to(self, host, access_mask=SnapAccessLevelEnum.READ_WRITE):

        # Use group snap id if it is a member snap.
        snap_id = self._get_group_snap_id()

        host_access = [{'host': host, 'allowedAccess': access_mask}]

        # Needs a update, otherwise the existing host may be overwritten.
        self.update()
        # If this snap is detached, attach it to the host.
        if not self.host_access:
            resp = self._cli.action(self.resource_class,
                                    snap_id, 'attach',
                                    hostAccess=self._cli.make_body(
                                        host_access))
        # Else, if this snap has been attached to other host,
        # don't overwrite it.
        else:
            host_access += [{'host': item.host,
                             'allowedAccess': item.allowed_access}
                            for item in filter(lambda x: x.host.id != host.id,
                                               self.host_access)]

            resp = self._cli.action(self.resource_class,
                                    snap_id, 'modify',
                                    hostAccess=self._cli.make_body(
                                        host_access))
        resp.raise_if_err()

        return resp

    def detach_from(self, host):
        # No need to pass host in to detach action of snap.
        # Snap host access will all be detached.

        # Use group snap id if it is a member snap.
        snap_id = self._get_group_snap_id()
        resp = self._cli.action(self.resource_class,
                                snap_id, 'detach')
        resp.raise_if_err()
        return resp

    def restore(self, backup=None, delete_backup=False):
        """Restore the snapshot to the associated storage resource.

        :param backup: name of the backup snapshot
        :param delete_backup: Whether to delete the backup snap after a
                              successful restore.
        """
        resp = self._cli.action(self.resource_class, self.get_id(),
                                'restore', copyName=backup)
        resp.raise_if_err()
        backup = resp.first_content['backup']
        backup_snap = UnitySnap(_id=backup['id'], cli=self._cli)

        if delete_backup:
            log.info("Deleting the backup snap {} as the restoration "
                     "succeeded.".format(backup['id']))
            backup_snap.delete()

        return backup_snap

    def is_cg_snap(self):
        cg_type = enums.StorageResourceTypeEnum.CONSISTENCY_GROUP
        return (self.storage_resource.type == cg_type and
                not self.is_member_snap())

    def get_member_snap(self, lun):
        members = UnitySnapList(cli=self._cli, snap_group=self, lun=lun)
        return members.first_item

    @version('>=4.2')
    def thin_clone(self, name, io_limit_policy=None, description=None):
        """Creates a new thin clone from this snapshot.
        .. note:: this snapshot should not enable Auto-Delete.
        """
        if self.is_member_snap():
            raise UnityCGMemberActionNotSupportError()

        if self.lun and not self.lun.is_thin_enabled:
            raise UnityThinCloneNotAllowedError()

        return TCHelper.thin_clone(self._cli, self, name, io_limit_policy,
                                   description)

    def is_member_snap(self):
        """Returns True if it is a member snap of cg snap."""
        return self.snap_group is not None

    def delete(self, async_mode=False, even_attached=False):
        """Deletes the snapshot.

        :param async_mode: whether to delete the snapshot in async mode.
        :param even_attached: whether to delete the snapshot even it is
            attached to hosts.
        """
        try:
            return super(UnitySnap, self).delete(async_mode=async_mode)
        except UnityDeleteAttachedSnapError:
            if even_attached:
                log.debug("Force delete the snapshot even if it is attached. "
                          "First detach the snapshot from hosts, then delete "
                          "again.")
                # Currently `detach_from` doesn't process `host` parameter.
                # It always detaches the snapshot from all hosts. So pass in
                # `None` here.
                self.detach_from(None)
                return super(UnitySnap, self).delete(async_mode=async_mode)
            else:
                raise


class UnitySnapList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnitySnap



# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import logging

from storops.lib.common import instance_cache
from storops.unity.resource import UnityResource, UnityResourceList, \
    UnityAttributeResource
from storops.exception import UnityEthernetPortMtuSizeNotSupportError
from storops.exception import UnityEthernetPortSpeedNotSupportError
from storops.unity.enums import EPSpeedValuesEnum, IOLimitPolicyTypeEnum
from storops.lib.converter import from_hour, from_minute
from storops.lib.version import version

__author__ = 'Jay Xu'

LOG = logging.getLogger(__name__)


class UnityIpPort(UnityResource):
    def set_mtu(self, mtu):
        port = self.get_physical_port()
        port.modify(mtu=mtu)

    @instance_cache
    @version('>=4.1')
    def is_link_aggregation(self):
        port = UnityLinkAggregation.get(self._cli, self.get_id())
        return port.existed

    @instance_cache  # noqa
    @version('<4.1')
    def is_link_aggregation(self):
        # Link aggregation is not supported before Falcon 4.1
        return False

    def get_physical_port(self):
        """Returns the link aggregation object or the ethernet port object."""
        obj = None
        if self.is_link_aggregation():
            obj = UnityLinkAggregation.get(self._cli, self.get_id())
        else:
            obj = UnityEthernetPort.get(self._cli, self.get_id())
        return obj

    @property
    def mtu(self):
        port = self.get_physical_port()
        return port.mtu


class UnityIpPortList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityIpPort


class UnityFcPort(UnityResource):
    pass


class UnityFcPortList(UnityResourceList):
    def __init__(self, cli=None, port_ids=None, **filters):
        super(UnityFcPortList, self).__init__(cli, **filters)
        self._port_ids = None
        self._set_filter(port_ids)

    def _set_filter(self, port_ids=None, **kwargs):
        self._port_ids = port_ids

    def _filter(self, fc_port):
        ret = True
        if self._port_ids is not None:
            ret &= fc_port.get_id() in self._port_ids
        return ret

    @classmethod
    def get_resource_class(cls):
        return UnityFcPort


class UnityIoModule(UnityResource):
    pass


class UnityIoModuleList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityIoModule


class UnityIoLimitPolicy(UnityResource):
    @classmethod
    def create(cls, cli, name, max_iops=None, max_kbps=None,
               policy_type=None, is_shared=None, description=None,
               max_iops_density=None, max_kbps_density=None,
               burst_rate=None, burst_time=None, burst_frequency=None):
        if policy_type is None:
            policy_type = IOLimitPolicyTypeEnum.ABSOLUTE
        if is_shared is None:
            is_shared = False

        rule = cli.make_body(name='{}_rule'.format(name),
                             maxIOPS=max_iops,
                             maxKBPS=max_kbps,
                             maxIOPSDensity=max_iops_density,
                             maxKBPSDensity=max_kbps_density,
                             burstRate=burst_rate,
                             burstTime=from_minute(burst_time),
                             burstFrequency=from_hour(burst_frequency))
        resp = cli.post(cls().resource_class,
                        name=name,
                        description=description,
                        isShared=is_shared,
                        type=policy_type,
                        ioLimitRules=[rule])
        resp.raise_if_err()
        return cls(_id=resp.resource_id, cli=cli)

    def modify(self, name=None, policy_type=None,
               is_paused=None, max_iops=None, max_kbps=None,
               max_iops_density=None, max_kbps_density=None,
               description=None, is_shared=None,
               burst_rate=None, burst_time=None, burst_frequency=None):
        # name = self.name if name is None else self.name
        rule = self._cli.make_body(maxIOPS=max_iops,
                                   maxKBPS=max_kbps,
                                   maxIOPSDensity=max_iops_density,
                                   maxKBPSDensity=max_kbps_density,
                                   burstRate=burst_rate,
                                   burstTime=from_minute(burst_time),
                                   burstFrequency=from_hour(burst_frequency),
                                   id=self.io_limit_rule_settings[0].get_id())
        resp = self._cli.modify(self.resource_class, self.get_id(),
                                name=name,
                                description=description,
                                isShared=is_shared,
                                type=policy_type,
                                ioLimitRulesToModify=[rule])
        resp.raise_if_err()
        return self

    def apply_to_storage(self, *lun_list):
        return self._cli.action(self.resource_class, self.get_id(),
                                'applyToStorage', luns=lun_list)

    def remove_from_storage(self, *lun_list):
        return self._cli.action(self.resource_class, self.get_id(),
                                'removeFromStorage', luns=lun_list)


class UnityIoLimitPolicyList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityIoLimitPolicy


class UnityIoLimitRule(UnityResource):
    pass


class UnityIoLimitRuleList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityIoLimitRule


class UnityIoLimitRuleSetting(UnityAttributeResource):
    pass


class UnityIoLimitRuleSettingList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityIoLimitRuleSetting


class UnityIscsiPortal(UnityResource):
    @classmethod
    def get_nested_properties(cls):
        return 'iscsi_node.name'

    @classmethod
    def create(cls, cli, ethernet_port, ip, netmask=None,
               v6_prefix_len=None, vlan=None, gateway=None):
        if not isinstance(ethernet_port, UnityEthernetPort):
            ethernet_port = UnityEthernetPort.get(cli, _id=ethernet_port)

        req_body = cli.make_body(
            ethernetPort=ethernet_port,
            ipAddress=ip,
            netmask=netmask,
            v6PrefixLength=v6_prefix_len,
            vlanId=vlan,
            gateway=gateway,
        )
        resp = cli.post(cls().resource_class, **req_body)
        resp.raise_if_err()
        return cls(cli=cli, _id=resp.resource_id)

    def modify(self, ip=None, netmask=None,
               v6_prefix_len=None, vlan=None, gateway=None):
        req_body = self._cli.make_body(
            ipAddress=ip,
            netmask=netmask,
            v6PrefixLength=v6_prefix_len,
            vlanId=vlan,
            gateway=gateway
        )
        resp = self._cli.modify(self.resource_class,
                                self.get_id(), **req_body)
        resp.raise_if_err()
        return resp


class UnityIscsiPortalList(UnityResourceList):
    def __init__(self, cli=None, port_ids=None, **filters):
        super(UnityIscsiPortalList, self).__init__(cli, **filters)
        self._port_ids = None
        self._set_filter(port_ids)

    def _set_filter(self, port_ids=None, **kwargs):
        self._port_ids = port_ids

    def _filter(self, iscsi_portal):
        ret = True
        if self._port_ids is not None:
            ret &= iscsi_portal.ethernet_port.get_id() in self._port_ids
        return ret

    @classmethod
    def get_resource_class(cls):
        return UnityIscsiPortal


class UnityEthernetPort(UnityResource):
    def modify(self, speed=None, mtu=None):
        speed = EPSpeedValuesEnum.parse(speed)
        self._modify(self, speed, mtu)
        peer = self.get_peer()
        if peer.existed:
            self._modify(peer, speed, mtu)

    def get_peer(self):
        peer_id = self._get_peer_id(self.get_id())
        return UnityEthernetPort(cli=self._cli, _id=peer_id)

    def _modify(self, port, speed, mtu):
        if speed is not None:
            if speed not in self.supported_speeds:
                raise UnityEthernetPortSpeedNotSupportError
            if speed == port.requested_speed:
                speed = None

        if mtu is not None:
            if mtu not in self.supported_mtus:
                raise UnityEthernetPortMtuSizeNotSupportError
            if mtu == port.requested_mtu:
                mtu = None

        if speed is None and mtu is None:
            return

        resp = self._cli.modify(self.resource_class,
                                port.get_id(),
                                speed=speed, mtuSize=mtu)
        resp.raise_if_err()

    @staticmethod
    def _get_peer_id(_id):
        if _id.startswith('spa'):
            return _id.replace('spa', 'spb')
        return _id.replace('spb', 'spa')

    @property
    def parent_storage_processor(self):
        # For ports on IO modules, there is no `parent_storage_processor`.
        # Set it manually.
        return self.storage_processor


class UnityEthernetPortList(UnityResourceList):
    def __init__(self, cli=None, port_ids=None, **filters):
        super(UnityEthernetPortList, self).__init__(cli, **filters)
        self._port_ids = None
        self._set_filter(port_ids)

    def _set_filter(self, port_ids=None, **kwargs):
        self._port_ids = port_ids

    def _filter(self, ethernet_port):
        ret = True
        if self._port_ids is not None:
            ret &= ethernet_port.get_id() in self._port_ids
        return ret

    @classmethod
    def get_resource_class(cls):
        return UnityEthernetPort


@version(">=4.1")
class UnityLinkAggregation(UnityResource):
    @classmethod
    def create(cls, cli, ports, mtu=None):
        resp = cli.post(cls().resource_class,
                        ports=ports,
                        mtuSize=mtu)
        resp.raise_if_err()
        return cls(_id=resp.resource_id, cli=cli)

    def modify(self, mtu=None, add_ports=None, remove_ports=None):
        if isinstance(add_ports, UnityEthernetPort):
            add_ports = [add_ports]
        if isinstance(remove_ports, UnityEthernetPort):
            remove_ports = [remove_ports]

        resp = self._cli.modify(self.resource_class,
                                self.get_id(),
                                mtuSize=mtu,
                                addPorts=add_ports,
                                removePorts=remove_ports)
        resp.raise_if_err()


@version(">=4.1")
class UnityLinkAggregationList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityLinkAggregation


class UnityIscsiNode(UnityResource):
    @classmethod
    def get_nested_properties(cls):
        return (
            'ethernet_port.connector_type',
            'ethernet_port.speed',
            'ethernet_port.supported_speeds',
            'ethernet_port.health',
        )


class UnityIscsiNodeList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityIscsiNode


class UnitySasPort(UnityResource):
    @classmethod
    def get_nested_properties(cls):
        return (
            'parentIOModule.name',
            'parent_storage_processor.name',
        )


class UnitySasPortList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnitySasPort



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from hamcrest import assert_that, equal_to


__author__ = 'Cedric Zhuang'


def test_create_fs_success(unity_gf):
    fs = _create_test_fs(unity_gf)
    assert_that(fs.existed, equal_to(True))


def _create_test_fs(unity_gf):
    server = unity_gf.nas_server
    name = unity_gf.add_fs_name()
    return unity_gf.pool.create_filesystem(server, name, 3 * 1024 ** 3)


def test_extend_fs_success(unity_gf):
    fs = _create_test_fs(unity_gf)
    _5g = 5 * 1024 ** 3
    resp = fs.extend(_5g)
    assert_that(resp.is_ok(), equal_to(True))

    fs.update()
    assert_that(fs.size_total, equal_to(_5g))


def test_create_delete_snapshot(unity_gf):
    fs = unity_gf.nfs_share.filesystem
    snap = fs.create_snap()
    assert_that(snap.existed, equal_to(True))
    snap.delete()



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import time
from unittest import TestCase

from hamcrest import equal_to, assert_that, raises, less_than

from storops.exception import VNXCredentialError
from storops.vnx.navi_command import NaviCommand
from storops_test.vnx.cli_mock import patch_cli

__author__ = 'Cedric Zhuang'


class NaviCommandTest(TestCase):
    @patch_cli
    def test_username_password(self):
        cmd = NaviCommand('a', 'a', 1)
        assert_that(' '.join(map(str, cmd.get_credentials())),
                    equal_to('-user a -password a -scope 1'))

    @patch_cli
    def test_default_security_file(self):
        cmd = NaviCommand()
        assert_that(cmd.get_credentials(), equal_to([]))

    @patch_cli
    def test_security_file(self):
        cmd = NaviCommand(sec_file=r'/a/b/c.key')
        assert_that(' '.join(cmd.get_credentials()),
                    equal_to('-secfilepath /a/b/c.key'))

    @patch_cli
    def test_security_file_precedence(self):
        cmd = NaviCommand(sec_file=r'/a/b/c.key',
                          username='admin', password='')
        assert_that(' '.join(cmd.get_credentials()),
                    equal_to('-secfilepath /a/b/c.key'))

    @patch_cli
    def test_username_password_timeout(self):
        cmd = NaviCommand('a', 'a', 1, timeout=20)
        assert_that(' '.join(map(str, cmd.get_credentials())),
                    equal_to('-user a -password a -scope 1 -t 20'))

    @patch_cli
    def test_sec_file_empty_string(self):
        cmd = NaviCommand('a', 'a', 1, '', timeout=20)
        assert_that(' '.join(map(str, cmd.get_credentials())),
                    equal_to('-user a -password a -scope 1 -t 20'))

    @patch_cli
    def test_timeout_max_and_min(self):
        cmd = NaviCommand()
        cmd._timeout = 0.5
        assert_that(cmd.timeout, equal_to(cmd.MIN_TIMEOUT))
        cmd._timeout = 10 ** 10
        assert_that(cmd.timeout, equal_to(cmd.MAX_TIMEOUT))

    @patch_cli(output='security_low.txt')
    def test_security_level_low(self):
        cmd = NaviCommand()
        assert_that(cmd.get_security_level('naviseccli'), equal_to('low'))

    def test_error_credentials(self):
        cmd = NaviCommand('a')
        assert_that(cmd.is_credential_valid, equal_to(True))
        assert_that(cmd.get_credentials, raises(VNXCredentialError, 'missing'))
        assert_that(cmd.is_credential_valid, equal_to(False))

    def test_timeout_error(self):
        cmd = NaviCommand()
        start = time.time()
        cmd.execute('python'.split(), timeout=0.1)
        dt = time.time() - start
        assert_that(dt, less_than(1))



# coding=utf-8
# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals, division

import inspect
import os
import sys

import six

from storops.lib.common import cache, all_not_none
from storops.lib.metric import CalculatorMetaInfo, MetricConfigParser, \
    MetricConfigList

__author__ = 'Cedric Zhuang'


class IdValues(object):
    NaN = float('nan')

    def __init__(self, data=None):
        if data is None:
            data = {}
        self._data = data

    def __sub__(self, other):
        if other is None:
            ret = self
        else:
            ret = self + (-other)
        return ret

    def __rsub__(self, other):
        if other is None:
            ret = -self
        else:
            ret = other + (-self)
        return ret

    def __neg__(self):
        return IdValues({k: -v for k, v in self._data.items()})

    @staticmethod
    def _add(op1, op2):
        if op1 is None:
            r = op2
        elif op2 is None:
            r = op1
        else:
            r = op1 + op2
        return r

    def __add__(self, other):
        if other is not None:
            defaults = 0
            op = self._add

            ret = self._apply_op(defaults, op, other)
        else:
            ret = self.copy()
        return ret

    def _apply_op(self, defaults, op, other):
        if isinstance(other, IdValues):
            keys = self.keys_union(other)
            ret = IdValues(
                {k: op(self.get(k, defaults), other.get(k, defaults))
                 for k in keys})
        else:
            ret = IdValues({k: op(v, other)
                            for k, v in self._data.items()})
        return ret

    def keys_union(self, other):
        return set(other.keys()).union(set(self.keys()))

    def __radd__(self, other):
        return self.__add__(other)

    def __truediv__(self, other):
        return self.__div__(other)

    @classmethod
    def _div(cls, op1, op2):
        if op1 == 0:
            r = 0
        elif op1 is None or op2 is None or op2 == 0:
            r = cls.NaN
        else:
            r = op1 / op2
        return r

    def __div__(self, other):
        return self._apply_op(self.NaN, self._div, other)

    def __rtruediv__(self, other):
        return IdValues({k: self._div(other, v)
                         for k, v in self._data.items()})

    def __rdiv__(self, other):
        return self.__rtruediv__(other)

    @staticmethod
    def _mul(op1, op2):
        if op1 is None:
            r = op2
        elif op2 is None:
            r = op1
        else:
            r = op1 * op2
        return r

    def __mul__(self, other):
        if other is not None:
            ret = self._apply_op(1, self._mul, other)
        else:
            ret = self.copy()
        return ret

    def __rmul__(self, other):
        return self.__mul__(other)

    def copy(self):
        return IdValues({k: v for k, v in self._data.items()})

    def __len__(self):
        return len(self._data)

    def __getitem__(self, item):
        return self._data.get(item, self.NaN)

    def keys(self):
        return self._data.keys()

    def get(self, k, default=None):
        return self._data.get(k, default)

    def set(self, k, v):
        self._data[k] = v

    def __str__(self):
        return '{}({})'.format(self.__class__.__name__, self._data)

    def __repr__(self):
        return str(self)


_BLOCK_SIZE = 512


def metric_calculator(f):
    def wrapper(path, prev, curr, obj_id=None):
        ret = f(path, prev, curr)
        if obj_id is not None:
            ret = ret[obj_id]
        return ret

    return wrapper


@metric_calculator
def delta_ps(path, prev, curr):
    path = only_one_path(path)

    if all_not_none(prev, curr):
        prev_counter = prev.by_path(path)
        curr_counter = curr.by_path(path)
    else:
        prev_counter = None
        curr_counter = None

    if all_not_none(prev_counter, curr_counter):
        delta = curr_counter.sum_sp() - prev_counter.sum_sp()
        ret = delta / curr.diff_seconds(prev)
    else:
        ret = IdValues()
    return ret


@metric_calculator
def total_delta_ps(path, prev, curr):
    if len(path) != 2:
        raise ValueError('takes in "reads" and "writes" counter.')

    read_path, write_path = path
    return delta_ps(read_path, prev, curr) + delta_ps(write_path, prev, curr)


def only_one_path(path):
    if isinstance(path, (list, tuple, set)):
        if len(path) != 1:
            raise ValueError('takes in one and only one path.')
        path = next(iter(path))
    return path


@metric_calculator
def mb_ps_by_block(path, prev, curr):
    return delta_ps(path, prev, curr) * _BLOCK_SIZE / (2.0 ** 20)


@metric_calculator
def mb_ps_by_byte(path, prev, curr):
    return delta_ps(path, prev, curr) / (2.0 ** 20)


@metric_calculator
def busy_idle_util(path, prev, curr):
    if len(path) != 2:
        raise ValueError('takes in "busy" and "idle" counter.')

    busy_path, idle_path = path
    if all_not_none(prev, curr):
        prev_busy = prev.by_path(busy_path)
        prev_idle = prev.by_path(idle_path)
        curr_busy = curr.by_path(busy_path)
        curr_idle = curr.by_path(idle_path)

        delta_busy = curr_busy.sum_sp() - prev_busy.sum_sp()
        delta_idle = curr_idle.sum_sp() - prev_idle.sum_sp()
        ret = delta_busy * 100 / (delta_idle + delta_busy)
    else:
        ret = IdValues()
    return ret


def _io_ps(path, prev, curr, attr):
    path = only_one_path(path)

    if all_not_none(prev, curr):
        prev_counter = prev.by_path(path)
        curr_counter = curr.by_path(path)
    else:
        prev_counter = None
        curr_counter = None

    if all_not_none(prev_counter, curr_counter):
        delta = getattr(curr_counter, attr) - getattr(prev_counter, attr)
        ret = delta / curr.diff_seconds(prev)
    else:
        ret = IdValues()
    return ret


@metric_calculator
def sp_delta_ps(path, prev, curr):
    attr = 'sp_values'
    return _io_ps(path, prev, curr, attr)


def _total_io_ps(path, prev, curr, err_msg, func):
    if len(path) != 2:
        raise ValueError(err_msg)

    part1, part2 = path
    return func(part1, prev, curr) + func(part2, prev, curr)


@metric_calculator
def sp_total_delta_ps(path, prev, curr):
    err_msg = 'takes in "reads" and "writes" counter.'
    return _total_io_ps(path, prev, curr, err_msg, sp_delta_ps)


@metric_calculator
def sp_mb_ps_by_byte(path, prev, curr):
    return sp_delta_ps(path, prev, curr) / (2.0 ** 20)


@metric_calculator
def sp_mb_ps_by_block(path, prev, curr):
    return sp_delta_ps(path, prev, curr) * _BLOCK_SIZE / (2.0 ** 20)


@metric_calculator
def sp_busy_idle_util(path, prev, curr):
    if len(path) != 2:
        raise ValueError('takes in "busy" and "idle" counter.')

    return _sp_pct(path, curr, prev)


@metric_calculator
def sp_hit_ratio(path, prev, curr):
    if len(path) != 2:
        raise ValueError('takes in "hit" and "miss" counter.')

    return _sp_pct(path, curr, prev)


@metric_calculator
def sp_fact(path, _, curr):
    path = only_one_path(path)
    return curr.by_path(path).sp_values


def _sp_pct(path, curr, prev):
    part1, part2 = path
    if all_not_none(prev, curr):
        prev_part1 = prev.by_path(part1)
        prev_part2 = prev.by_path(part2)
        curr_part1 = curr.by_path(part1)
        curr_part2 = curr.by_path(part2)

        delta_part1 = curr_part1.sp_values - prev_part1.sp_values
        delta_part2 = curr_part2.sp_values - prev_part2.sp_values
        ret = delta_part1 * 100 / (delta_part2 + delta_part1)
    else:
        ret = IdValues()
    return ret


@metric_calculator
def sp_sum_values(path, _, curr):
    path = only_one_path(path)
    return curr.by_path(path).sp_sum_values


@metric_calculator
def sp_io_rate(path, prev, curr):
    attr = 'sp_sum_values'
    return _io_ps(path, prev, curr, attr)


@metric_calculator
def system_delta_ps(path, prev, curr):
    attr = 'sum_sp_values'
    return _io_ps(path, prev, curr, attr)


@metric_calculator
def system_total_delta_ps(path, prev, curr):
    err_msg = 'takes in "reads" and "writes" counter.'
    return _total_io_ps(path, prev, curr, err_msg, system_delta_ps)


@metric_calculator
def disk_response_time(path, prev, curr):
    if len(path) != 5:
        raise ValueError(
            'takes in "busyTicks", "sumArrivalQueueLength", "reads", '
            '"writes" and "coreCount" counter.')

    busy_tick_path, sum_arr_ql_path, read_path, write_path, core_count_path = \
        path
    if all_not_none(prev, curr):
        prev_busy_tick = prev.by_path(busy_tick_path)
        prev_sum_arr_ql = prev.by_path(sum_arr_ql_path)
        prev_read = prev.by_path(read_path)
        prev_write = prev.by_path(write_path)
        prev_core_count = prev.by_path(core_count_path)

        curr_busy_tick = curr.by_path(busy_tick_path)
        curr_sum_arr_ql = curr.by_path(sum_arr_ql_path)
        curr_read = curr.by_path(read_path)
        curr_write = curr.by_path(write_path)
        curr_core_count = curr.by_path(core_count_path)

        delta_busy_tick = curr_busy_tick.sum_sp() - prev_busy_tick.sum_sp()
        delta_sum_arr_ql = curr_sum_arr_ql.sum_sp() - prev_sum_arr_ql.sum_sp()
        delta_read = curr_read.sum_sp() - prev_read.sum_sp()
        delta_write = curr_write.sum_sp() - prev_write.sum_sp()

        delta_read_core_count = curr_read.combine_numeric_values(
            curr_core_count) - prev_read.combine_numeric_values(
            prev_core_count)
        delta_write_core_count = curr_write.combine_numeric_values(
            curr_core_count) - prev_write.combine_numeric_values(
            prev_core_count)

        ret = delta_busy_tick * delta_sum_arr_ql / (
            (delta_read + delta_write) *
            (delta_read_core_count + delta_write_core_count))

    else:
        ret = IdValues()

    return ret


def _queue_length(path, prev, curr, err_msg, func):
    if len(path) != 3:
        raise ValueError(err_msg)

    part1, part2, part3 = path
    if all_not_none(prev, curr):
        prev_part1 = prev.by_path(part1)
        prev_part2 = prev.by_path(part2)
        prev_part3 = prev.by_path(part3)

        curr_part1 = curr.by_path(part1)
        curr_part2 = curr.by_path(part2)
        curr_part3 = curr.by_path(part3)

        delta_part1 = curr_part1.sum_sp() - prev_part1.sum_sp()
        delta_part2 = curr_part2.sum_sp() - prev_part2.sum_sp()
        delta_part3 = curr_part3.sum_sp() - prev_part3.sum_sp()

        ret = func(delta_part1, delta_part2, delta_part3)

    else:
        ret = IdValues()

    return ret


@metric_calculator
def disk_queue_length(path, prev, curr):
    err_msg = 'takes in "sumArrivalQueueLength", "reads" and "writes" counter.'
    return _queue_length(path, prev, curr, err_msg,
                         lambda x, y, z: x / (y + z))


@metric_calculator
def lun_response_time(path, prev, curr):
    err_msg = 'takes in "totalIoTime", "reads" and "writes" counter.'
    return _queue_length(path, prev, curr, err_msg,
                         lambda x, y, z: x / (y + z))


@metric_calculator
def lun_queue_length(path, prev, curr):
    err_msg = 'takes in "currentIOCount", "busyTime" and "idleTime" counter.'
    return _queue_length(path, prev, curr, err_msg,
                         lambda x, y, z: x * y / (y + z))


def _byte_rate(path, prev, curr, err_msg, attr):
    if len(path) != 2:
        raise ValueError(err_msg)

    part1, part2 = path
    if all_not_none(prev, curr):
        delta = getattr(curr.by_path(part1), attr)(
            curr.by_path(part2)) - getattr(prev.by_path(part1), attr)(
            prev.by_path(part2))
        ret = delta / curr.diff_seconds(prev)
    else:
        ret = IdValues()

    return ret


@metric_calculator
def byte_rate(path, prev, curr):
    err_msg = 'takes in "Blocks" and "blockSize" counter.'
    attr = 'combine_numeric_values'
    return _byte_rate(path, prev, curr, err_msg, attr)


@metric_calculator
def sp_byte_rate(path, prev, curr):
    err_msg = 'takes in "Blocks" and "blockSize" counter.'
    attr = 'combine_sp_values'
    return _byte_rate(path, prev, curr, err_msg, attr)


@metric_calculator
def system_byte_rate(path, prev, curr):
    err_msg = 'takes in "Blocks" and "blockSize" counter.'
    attr = 'sum_combined_sp_values'
    return _byte_rate(path, prev, curr, err_msg, attr)


def _total_byte_rate(path, prev, curr, err_msg, func):
    if len(path) != 3:
        raise ValueError(err_msg)

    part1, part2, part3 = path
    return func((part1, part3), prev, curr) + func((part2, part3), prev, curr)


@metric_calculator
def total_byte_rate(path, prev, curr):
    err_msg = 'takes in "readBlocks", "writeBlocks" and "blockSize" counter.'
    return _total_byte_rate(path, prev, curr, err_msg, byte_rate)


@metric_calculator
def sp_total_byte_rate(path, prev, curr):
    err_msg = 'takes in "readBlocks", "writeBlocks" and "blockSize" counter.'
    return _total_byte_rate(path, prev, curr, err_msg, sp_byte_rate)


@metric_calculator
def system_total_byte_rate(path, prev, curr):
    err_msg = 'takes in "readBlocks", "writeBlocks" and "blockSize" counter.'
    return _total_byte_rate(path, prev, curr, err_msg, system_byte_rate)


@cache
def _module_functions():
    return dict(inspect.getmembers(sys.modules[__name__]))


class UnityMetricConfig(object):
    def __init__(self, config):
        self.name = config.get('name')
        paths = self._get_paths(config)
        self.paths = paths
        self.calculator = self._get_calculator(config)

    @staticmethod
    def _get_calculator(config):
        calc_name = config.get('calculator')
        if calc_name is None:
            ret = delta_ps
        else:
            ret = _module_functions().get(calc_name)
        return ret

    @staticmethod
    def _get_paths(config):
        paths = config.get('paths')
        if isinstance(paths, six.string_types):
            paths = [paths]
        return paths


class UnityMetricConfigList(MetricConfigList):
    def init_metric_config(self, raw_config):
        return UnityMetricConfig(raw_config)


class UnityMetricConfigParser(MetricConfigParser):
    @classmethod
    def get_folder(cls):
        return os.path.dirname(inspect.getfile(cls))

    @classmethod
    def get_config(cls, name):
        name = cls._get_clz_name(name)
        return UnityMetricConfigList(cls._read_configs().get(name))

    @classmethod
    @cache
    def paths(cls, clz_list=None):
        if clz_list is None:
            clz_list = cls._read_configs().keys()

        return [path for config in map(cls.get_config, clz_list)
                for path in config.paths()]


class UnityCalculatorMetaInfo(CalculatorMetaInfo):
    def get_config_parser(self):
        return UnityMetricConfigParser()

    def get_metric_value(self, clz, metric_name, cli, obj=None):
        if not hasattr(cli, 'curr_counter'):
            raise ValueError('cli should has "curr_counter" attribute.')
        if not hasattr(cli, 'prev_counter'):
            raise ValueError('cli should has "prev_counter" attribute.')

        config = self.get_config(clz).get_metric_config(metric_name)
        return config.calculator(config.paths, cli.prev_counter,
                                 cli.curr_counter, obj)

    def get_all_paths(self, clz_list=None):
        if clz_list is not None:
            clz_list = tuple(clz_list)
        return self._metric_config.paths(clz_list)


calculators = UnityCalculatorMetaInfo()



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, raises, equal_to, has_item, none, is_not, \
    only_contains, contains_string

from storops.exception import VNXInvalidCliParamError, \
    VNXPortNotInitializedError, VNXInitiatorExistedError, \
    VNXDeleteHbaNotFoundError, VNXPingNodeTimeOutError, VNXGateWayError, \
    VNXVirtualPortNotFoundError
from storops.lib.common import instance_cache
from storops.vnx.enums import VNXSPEnum, VNXPortType
from storops.vnx.resource.port import VNXHbaPort, VNXSPPort, \
    VNXConnectionPort, VNXStorageGroupHBA
from storops.vnx.resource.sg import VNXStorageGroup
from storops_test.vnx.cli_mock import patch_cli, t_cli
from storops_test.vnx.resource.fakes import STORAGE_GROUP_HBA

__author__ = 'Cedric Zhuang'


class VNXSPPortTest(TestCase):
    @patch_cli
    def test_port_list(self):
        ports = VNXSPPort.get(t_cli())
        assert_that(len(ports), equal_to(32))

    @patch_cli
    def test_port_get_sp(self):
        ports = VNXSPPort.get(t_cli(), VNXSPEnum.SP_B)
        assert_that(len(ports), equal_to(16))

    @patch_cli
    def test_port_get_id(self):
        ports = VNXSPPort.get(t_cli(), port_id=5)
        assert_that(len(ports), equal_to(2))

    @patch_cli
    def test_index_sequence(self):
        # this test will fail if the index is not used as splitter is wrong
        ports = VNXSPPort.get(t_cli(), VNXSPEnum.SP_A, 15)
        assert_that(ports[0].wwn, equal_to(
            '50:06:01:60:B6:E0:16:81:50:06:01:67:36:E4:16:81'))

    @patch_cli
    def test_get_port_property(self):
        port = VNXSPPort.get(t_cli(), VNXSPEnum.SP_A, 0)[0]
        assert_that(port.sp, equal_to(VNXSPEnum.SP_A))
        assert_that(port.port_id, equal_to(0))
        assert_that(port.index, equal_to('A_0'))
        assert_that(port.wwn, equal_to(
            '50:06:01:60:B6:E0:16:81:50:06:01:60:36:E0:16:81'))
        assert_that(port.link_status, equal_to('Up'))
        assert_that(port.port_status, equal_to('Online'))
        assert_that(port.switch_present, equal_to(True))
        assert_that(port.speed_value, equal_to('8Gbps'))
        assert_that(port.registered_initiators, equal_to(3))
        assert_that(port.logged_in_initiators, equal_to(1))
        assert_that(port.not_logged_in_initiators, equal_to(2))
        assert_that(port.type, equal_to(VNXPortType.FC))
        assert_that(port.display_name, equal_to('A-0'))
        assert_that(port.sfp_state, equal_to('Online'))
        assert_that(port.reads, equal_to(1))
        assert_that(port.writes, equal_to(2))
        assert_that(port.blocks_read, equal_to(3))
        assert_that(port.blocks_written, equal_to(4))
        assert_that(port.queue_full_busy, equal_to(5))
        assert_that(port.i_o_module_slot, equal_to('6'))
        assert_that(port.physical_port_id, equal_to(7))
        assert_that(port.usage, equal_to("General"))
        assert_that(port.sfp_connector_emc_part_number,
                    equal_to('019-078-042'))
        assert_that(port.sfp_connector_emc_serial_number,
                    equal_to('000000000000000'))
        assert_that(port.sfp_connector_vendor_part_number, equal_to('N/A'))
        assert_that(port.sfp_connector_vendor_serial_number,
                    equal_to('PL31D5E'))

    @patch_cli
    def test_get_port_by_type(self):
        ports = VNXSPPort.get(cli=t_cli(), port_type=VNXPortType.ISCSI)
        assert_that(len(ports), equal_to(4))
        ports = VNXSPPort.get(cli=t_cli(), port_type=VNXPortType.FC)
        assert_that(len(ports), equal_to(28))

    @property
    @instance_cache
    def port_a0(self):
        return VNXSPPort.get(t_cli(), sp=VNXSPEnum.SP_A, port_id=0)[0]

    @patch_cli
    def test_port_read_iops(self):
        assert_that(self.port_a0.read_iops, equal_to(10.0))

    @patch_cli
    def test_port_write_iops(self):
        assert_that(self.port_a0.write_iops, equal_to(20.0))

    @patch_cli
    def test_port_total_iops(self):
        assert_that(self.port_a0.total_iops, equal_to(30.0))

    @patch_cli
    def test_port_read_mbps(self):
        assert_that(self.port_a0.read_mbps, equal_to(1.0))

    @patch_cli
    def test_port_write_mbps(self):
        assert_that(self.port_a0.write_mbps, equal_to(1.5))

    @patch_cli
    def test_port_total_mbps(self):
        assert_that(self.port_a0.total_mbps, equal_to(2.5))

    @patch_cli
    def test_port_read_size_kb(self):
        assert_that(self.port_a0.read_size_kb, equal_to(102.4))

    @patch_cli
    def test_port_write_size_kb(self):
        assert_that(self.port_a0.write_size_kb, equal_to(76.8))

    @patch_cli
    def test_property_names(self):
        assert_that(self.port_a0.property_names(), has_item('read_iops'))


class VNXHbaPortTest(TestCase):
    def test_from_storage_group_hba(self):
        hba = VNXStorageGroupHBA.parse(STORAGE_GROUP_HBA)
        port = VNXHbaPort.from_storage_group_hba(hba)
        assert_that(port.sp, equal_to(VNXSPEnum.SP_A))
        assert_that(port.port_id, equal_to(3))
        assert_that(port.vport_id, equal_to(1))
        assert_that(port.type, equal_to(VNXPortType.ISCSI))
        assert_that(port.host_initiator_list,
                    has_item('iqn.1991-05.com.microsoft:abc.def.dev'))

    def test_hash(self):
        ports = {
            VNXHbaPort.create(VNXSPEnum.SP_A, 1),
            VNXHbaPort.create(VNXSPEnum.SP_B, 1),
            VNXHbaPort.create(VNXSPEnum.SP_A, 1)
        }
        assert_that(len(ports), equal_to(2))

    def test_set_sp(self):
        port = VNXHbaPort.create('A', 3)
        assert_that(port.sp, equal_to(VNXSPEnum.SP_A))

    def test_set_sp_error(self):
        port = VNXHbaPort.create('Z', 3)
        assert_that(port.is_valid(), equal_to(False))
        assert_that(port.sp, none())

    def test_set_number_error(self):
        def f():
            port = VNXHbaPort.create('A', 'a1')
            assert_that(port.is_valid(), equal_to(False))
            assert_that(port.port_id, none())

        assert_that(f, raises(ValueError, 'must be an integer.'))

    def test_create_tuple_input(self):
        inputs = ('a', 5)
        port = VNXHbaPort.create(*inputs)
        assert_that(port.sp, equal_to(VNXSPEnum.SP_A))
        assert_that(port.port_id, equal_to(5))

    def test_get_sp_index(self):
        port = VNXHbaPort.create('spb', '5')
        assert_that(port.get_sp_index(), equal_to('b'))
        assert_that(port.port_id, equal_to(5))

    def test_equal(self):
        spa_1 = VNXHbaPort.create(VNXSPEnum.SP_A, 1)
        spa_1_dup = VNXHbaPort.create(VNXSPEnum.SP_A, 1)
        spa_2 = VNXHbaPort.create(VNXSPEnum.SP_A, 2)
        assert_that(spa_1_dup, equal_to(spa_1))
        assert_that(spa_1, is_not(equal_to(spa_2)))

    def test_as_tuple(self):
        port = VNXHbaPort.create(VNXSPEnum.SP_A, 1)
        assert_that(port.as_tuple(), only_contains(VNXSPEnum.SP_A, 1))

    def test_repr(self):
        port = VNXHbaPort.create(VNXSPEnum.SP_B, 3)
        ret = port.__repr__()
        assert_that(ret, contains_string('"sp": "VNXSPEnum.SP_B"'))
        assert_that(ret, contains_string('"existed": true'))


class VNXConnectionPortTest(TestCase):
    def test_port(self):
        return VNXConnectionPort(sp='a', port_id=4, cli=t_cli())

    @patch_cli
    def test_properties(self):
        port = self.test_port()
        assert_that(port.sp, equal_to(VNXSPEnum.SP_A))
        assert_that(port.port_id, equal_to(4))
        assert_that(port.wwn,
                    equal_to('iqn.1992-04.com.emc:cx.apm00153906536.a4'))
        assert_that(port.iscsi_alias, equal_to('6536.a4'))
        assert_that(port.enode_mac_address, equal_to('00-60-16-45-5D-FC'))
        assert_that(port.virtual_port_id, equal_to(0))
        assert_that(port.vport_id, equal_to(0))
        assert_that(port.vlan_id, none())
        assert_that(port.current_mtu, equal_to(1500))
        assert_that(port.auto_negotiate, equal_to(False))
        assert_that(port.port_speed, equal_to('10000 Mb'))
        assert_that(port.host_window, equal_to('256K'))
        assert_that(port.replication_window, equal_to('256K'))
        assert_that(port.ip_address, equal_to('192.168.4.52'))
        assert_that(port.subnet_mask, equal_to('255.255.255.0'))
        assert_that(port.gateway_address, equal_to('0.0.0.0'))
        assert_that(port.type, equal_to(VNXPortType.ISCSI))
        assert_that(port.existed, equal_to(True))
        assert_that(port.display_name, equal_to('A-4-0'))

    @patch_cli
    def test_get_all(self):
        ports = VNXConnectionPort.get(t_cli())
        assert_that(len(ports), equal_to(23))

    @patch_cli
    def test_get_by_sp(self):
        ports = VNXConnectionPort.get(t_cli(), VNXSPEnum.SP_A)
        assert_that(len(ports), equal_to(13))

    @patch_cli
    def test_get_by_port(self):
        ports = VNXConnectionPort.get(t_cli(), port_id=8)
        assert_that(len(ports), equal_to(2))

    @patch_cli
    def test_get_by_type(self):
        ports = VNXConnectionPort.get(t_cli(), port_type=VNXPortType.ISCSI)
        assert_that(len(ports), equal_to(19))
        ports = VNXConnectionPort.get(t_cli(), port_type=VNXPortType.FCOE)
        assert_that(len(ports), equal_to(4))

    @patch_cli
    def test_get_by_sp_port_id(self):
        ports = VNXConnectionPort.get(t_cli(), VNXSPEnum.SP_A, 4)
        assert_that(len(ports), equal_to(1))
        port = ports[0]
        assert_that(port.port_id, equal_to(4))
        assert_that(port.sp, equal_to(VNXSPEnum.SP_A))
        assert_that(port.virtual_port_id, equal_to(0))
        assert_that(port.vport_id, equal_to(0))

    @patch_cli
    def test_get_single(self):
        port = VNXConnectionPort.get(t_cli(), VNXSPEnum.SP_A, 5, 2)
        assert_that(port.port_id, equal_to(5))
        assert_that(port.sp, equal_to(VNXSPEnum.SP_A))
        assert_that(port.virtual_port_id, equal_to(2))
        assert_that(port.vport_id, equal_to(2))

    @patch_cli
    def test_get_port_not_found(self):
        ports = VNXConnectionPort.get(t_cli(), VNXSPEnum.SP_A, 44)
        assert_that(len(ports), equal_to(0))

    @patch_cli
    def test_get_port_with_vport_not_found(self):
        port = VNXConnectionPort.get(t_cli(), VNXSPEnum.SP_B, 10, 0)
        assert_that(port.existed, equal_to(False))

    @patch_cli
    def test_delete_fc_hba_success(self):
        uid = '00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00'
        # no error raised
        VNXSPPort.delete_hba(t_cli(), uid)

    @patch_cli
    def test_delete_hba_already_removed(self):
        def f():
            uid = '00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:01'
            VNXSPPort.delete_hba(t_cli(), uid)

        assert_that(f, raises(VNXDeleteHbaNotFoundError))

    @patch_cli
    def test_ping_node_timeout(self):
        def f():
            port = VNXConnectionPort.get(t_cli(), VNXSPEnum.SP_A, 8)[0]
            port.ping_node('10.244.211.3', count=1)

        assert_that(f, raises(VNXPingNodeTimeOutError))

    @patch_cli
    def test_ping_node_success(self):
        port = VNXConnectionPort.get(t_cli(), VNXSPEnum.SP_A, 8)[0]
        # success, no error raised
        port.ping_node('10.244.211.4', count=1)

    @patch_cli
    def test_ping_node_multiple_success(self):
        port = VNXConnectionPort.get(t_cli(), VNXSPEnum.SP_A, 8)[0]
        # success, no error raised
        port.ping_node('10.244.211.5')

    @patch_cli
    def test_config_ip_gateway_error(self):
        def f():
            port = VNXConnectionPort.get(t_cli(), VNXSPEnum.SP_A, 10)[0]
            port.config_ip('6.6.6.6', '255.255.255.0', '5.5.5.1')

        assert_that(f, raises(VNXGateWayError, 'netmask'))

    @patch_cli
    def test_delete_ip_not_found(self):
        def f():
            port = VNXConnectionPort.get(t_cli(), VNXSPEnum.SP_A, 10)[0]
            port.delete_ip()

        assert_that(f, raises(VNXVirtualPortNotFoundError, 'not found'))

    @patch_cli
    def test_port_display_name(self):
        port = VNXConnectionPort.get(t_cli(), VNXSPEnum.SP_A, 5, 0)
        assert_that('A-10-0', port.display_name)


def test_hba():
    return VNXStorageGroupHBA().update(STORAGE_GROUP_HBA)


class VNXStorageGroupHBATest(TestCase):
    def test_properties(self):
        hba = test_hba()
        assert_that(hba.host_name, equal_to('abc.def.dev'))
        assert_that(hba.initiator_ip, equal_to('10.244.209.72'))
        assert_that(hba.sp_port, equal_to('A-3v1'))

    def test_sp(self):
        assert_that(test_hba().sp, equal_to(VNXSPEnum.SP_A))

    def test_uid(self):
        assert_that(test_hba().uid,
                    equal_to('iqn.1991-05.com.microsoft:abc.def.dev'))

    def test_port_id(self):
        assert_that(test_hba().port_id, equal_to(3))

    def test_vport_id(self):
        assert_that(test_hba().vport_id, equal_to(1))

    def test_port_type(self):
        assert_that(test_hba().port_type,
                    equal_to(VNXPortType.ISCSI))

    @patch_cli
    def test_set_path_with_sp_port_invalid_wwn(self):
        def f():
            ports = VNXSPPort.get(sp=VNXSPEnum.SP_A, port_id=0, cli=t_cli())
            sg = VNXStorageGroup(cli=t_cli(), name='sg0')
            sg.set_path(ports[0], '11:22:33', 'host0')

        assert_that(f, raises(VNXInvalidCliParamError))

    @patch_cli
    def test_set_path_with_fc_port_success(self):
        wwn = '01:02:03:04:05:06:07:08:09:0A:0B:0C:0D:0E:0F:10'
        ports = VNXSPPort.get(sp=VNXSPEnum.SP_A, port_id=0, cli=t_cli())
        sg = VNXStorageGroup(cli=t_cli(), name='sg0')
        # no exception
        sg.set_path(ports[0], wwn, 'host0')

    @patch_cli
    def test_set_path_with_iscsi_port_not_initialized(self):
        def f():
            uid = 'iqn.1992-04.com.abc:a.b.c'
            port = VNXConnectionPort.get(sp=VNXSPEnum.SP_A, port_id=10,
                                         cli=t_cli())[0]
            sg = VNXStorageGroup(cli=t_cli(), name='sg0')
            sg.set_path(port, uid, 'host0')

        assert_that(f, raises(VNXPortNotInitializedError))

    @patch_cli
    def test_set_path_with_fcoe_port_success(self):
        uid = 'iqn.1992-04.com.abc:a.b.c'
        port = VNXConnectionPort.get(sp=VNXSPEnum.SP_A, port_id=8,
                                     vport_id=0, cli=t_cli())
        sg = VNXStorageGroup(cli=t_cli(), name='sg0')
        # no error raised
        sg.connect_hba(port, uid, 'host0')

    @patch_cli
    def test_set_path_with_fcoe_already_existed(self):
        def f():
            uid = 'iqn.1992-04.com.abc:a.b.d'
            port = VNXConnectionPort.get(sp=VNXSPEnum.SP_A, port_id=8,
                                         vport_id=0, cli=t_cli())
            sg = VNXStorageGroup(cli=t_cli(), name='sg0')
            sg.set_path(port, uid, 'host0')

        assert_that(f, raises(VNXInitiatorExistedError))


class VNXPortTest(TestCase):
    def hba_port_set(self):
        return {
            VNXHbaPort.create(VNXSPEnum.SP_A, 1),
            VNXHbaPort.create(VNXSPEnum.SP_B, 1),
            VNXHbaPort.create(VNXSPEnum.SP_A, 4),
            VNXHbaPort.create(VNXSPEnum.SP_A, 3, vport_id=1),
            VNXHbaPort.create(VNXSPEnum.SP_A, 6, vport_id=0),
            VNXHbaPort.create(VNXSPEnum.SP_B, 4, vport_id=0),
        }

    @patch_cli
    def test_connection_port_not_equal_sp_port(self):
        c_port = VNXConnectionPort(sp='a', port_id=4, cli=t_cli())
        s_port = VNXSPPort.get(t_cli(), VNXSPEnum.SP_A, 4)
        # one has vport_id, others not, not equal
        assert_that(c_port, is_not(equal_to(s_port)))

    @patch_cli
    def test_connection_port_equal_sp_port(self):
        c_port = VNXConnectionPort(sp='a', port_id=9, cli=t_cli())
        s_port = VNXSPPort.get(t_cli(), VNXSPEnum.SP_A, 9)[0]
        assert_that(c_port, equal_to(s_port))

    @patch_cli
    def test_connection_port_equal_hba_port(self):
        c_port = VNXConnectionPort(sp='a', port_id=4, cli=t_cli())
        h_port = VNXHbaPort.create('a', 4, vport_id=0)
        assert_that(c_port, equal_to(h_port))

    @patch_cli
    def test_sp_port_in_hba_port_set(self):
        ports = self.hba_port_set()
        s_port = VNXSPPort.get(t_cli(), VNXSPEnum.SP_A, 4)[0]
        assert_that(ports, has_item(s_port))

    @patch_cli
    def test_connection_port_not_in_hba_port_set(self):
        ports = self.hba_port_set()
        c_port = VNXConnectionPort(sp='a', port_id=4, cli=t_cli())
        assert_that(ports, is_not(has_item(c_port)))

        c_port = VNXConnectionPort(sp='a', port_id=9, cli=t_cli())
        assert_that(ports, is_not(has_item(c_port)))

    @patch_cli
    def test_connection_port_in_hba_port_set(self):
        ports = self.hba_port_set()
        c_port = VNXConnectionPort(sp='a', port_id=6, vport_id=0, cli=t_cli())
        assert_that(ports, has_item(c_port))

    @patch_cli
    def test_connection_port_in_sp_port_list(self):
        c_port = VNXConnectionPort(sp='a', port_id=9, cli=t_cli())
        ports = VNXSPPort.get(t_cli())
        assert_that(ports, has_item(c_port))

    @patch_cli
    def test_hba_equal_connection_port(self):
        hba = test_hba()
        c_port = VNXConnectionPort(sp='a', port_id=3, vport_id=1, cli=t_cli())
        assert_that(hba, equal_to(c_port))

    @patch_cli
    def test_hba_not_equal_sp_port(self):
        hba = test_hba()
        s_port = VNXSPPort.get(t_cli(), VNXSPEnum.SP_A, 3)
        assert_that(hba, is_not(equal_to(s_port)))

    @patch_cli
    def test_hba_in_sp_port(self):
        sg = VNXStorageGroup(name='server7', cli=t_cli())
        hba = None
        for hba in sg.ports:
            if hba.sp == VNXSPEnum.SP_A and hba.port_id == 0:
                break
        ports = VNXSPPort.get(t_cli())
        assert_that(ports, has_item(hba))

    @patch_cli
    def test_hba_in_hba_port(self):
        hba = test_hba()
        ports = self.hba_port_set()
        assert_that(ports, has_item(hba))

    @patch_cli
    def test_hba_equal_hba_port(self):
        hba = test_hba()
        h_port = VNXHbaPort.create('a', 3, vport_id=1)
        assert_that(hba, equal_to(h_port))

    @patch_cli
    def test_hba_port_with_connection_port(self):
        sg = VNXStorageGroup.get(cli=t_cli(), name='microsoft')
        h_port = sg.get_ports(
            initiator_uid='iqn.1991-05.com.microsoft:abc.def.dev')
        c_port = VNXConnectionPort.get(
            sp=VNXSPEnum.SP_A, port_id=1, cli=t_cli())
        assert_that(len(h_port), equal_to(2))
        r = set(h_port) - {item for item in c_port}
        assert_that(len(r), equal_to(1))
        assert_that(list(r)[0].port_id, equal_to(1))
        assert_that(list(r)[0].sp, equal_to(VNXSPEnum.SP_B))

    @patch_cli
    def test_hba_port_with_sp_port(self):
        sg = VNXStorageGroup.get(cli=t_cli(), name='server7')
        sg_ports = sg.get_ports(
            '20:00:00:90:FA:53:4C:D0:10:00:00:90:FA:53:4C:D0')
        sp_ports = VNXSPPort.get(cli=t_cli(), sp=VNXSPEnum.SP_A,
                                 port_type=VNXPortType.FC)
        r = set(sg_ports) - set(sp_ports)
        for x in r:
            assert_that(x.sp, equal_to(VNXSPEnum.SP_B))

    @patch_cli
    def test_get_metrics_csv(self):
        ports = VNXSPPort.get(t_cli())
        csv = ports.get_metrics_csv()
        assert_that(csv, contains_string(',A_0,'))
        assert_that(csv, contains_string('30.0,2.5,'))
        assert_that(csv, contains_string('timestamp,name,'))



# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import logging

from storops.exception import UnityShareTypeNotSupportAccessControlError, \
    UnityHostNotFoundException, UnityCreateSnapError
from storops.lib.common import instance_cache
from storops.unity.enums import NFSShareDefaultAccessEnum, NFSTypeEnum, \
    NFSShareSecurityEnum, FilesystemSnapAccessTypeEnum
import storops.unity.resource.filesystem
import storops.unity.resource.snap
from storops.unity.resource import UnityResource, UnityResourceList
import storops.unity.resource.host
from storops.unity.resp import RestResponse

__author__ = 'Jay Xu'

LOG = logging.getLogger(__name__)


class UnityNfsHostConfig(object):
    def __init__(self, root=None, ro=None, rw=None, no_access=None,
                 nfs_share=None):
        if nfs_share is not None:
            root = nfs_share.root_access_hosts
            ro = nfs_share.read_only_hosts
            rw = nfs_share.read_write_hosts
            no_access = nfs_share.no_access_hosts

        self.root = root
        self.ro = ro
        self.rw = rw
        self.no_access = no_access

    @classmethod
    def _add(cls, left, right):
        if left is None:
            ret = right
        elif right is None:
            ret = left
        else:
            ret = left
            for r in right:
                if r not in left:
                    ret.append(r)
        return ret

    @classmethod
    def _delete(cls, left, right):
        if left is None:
            ret = None
        elif right is None:
            ret = left
        else:
            ret = []
            for l in left:
                if l not in right:
                    ret.append(l)
        return ret

    def allow_root(self, *hosts):
        self.delete_access(*hosts)
        self.root = self._add(self.root, hosts)
        return self

    def allow_ro(self, *hosts):
        self.delete_access(*hosts)
        self.ro = self._add(self.ro, hosts)
        return self

    def allow_rw(self, *hosts):
        self.delete_access(*hosts)
        self.rw = self._add(self.rw, hosts)
        self.root = self._add(self.root, hosts)
        return self

    def deny_access(self, *hosts):
        self.delete_access(*hosts)
        self.no_access = self._add(self.no_access, hosts)
        return self

    def delete_access(self, *hosts):
        self.rw = self._delete(self.rw, hosts)
        self.ro = self._delete(self.ro, hosts)
        self.no_access = self._delete(self.no_access, hosts)
        self.root = self._delete(self.root, hosts)
        return self

    @staticmethod
    def _inter(left, right):
        if left is None:
            left = []
        right_ids = [host.get_id() for host in right]
        return [host for host in left if host.get_id() in right_ids]

    def clear_all(self, *white_list_hosts):
        self.rw = self._inter(self.rw, white_list_hosts)
        self.ro = self._inter(self.ro, white_list_hosts)
        self.no_access = self._inter(self.no_access, white_list_hosts)
        self.root = self._inter(self.root, white_list_hosts)
        return self


class UnityNfsShare(UnityResource):
    @classmethod
    def create(cls, cli, name, fs, path=None, share_access=None):
        fs_clz = storops.unity.resource.filesystem.UnityFileSystem
        fs = fs_clz.get(cli, fs).verify()
        NFSShareDefaultAccessEnum.verify(share_access)
        sr = fs.storage_resource

        if path is None:
            path = '/'

        share_param = cli.make_body(defaultAccess=share_access)
        param = cli.make_body(name=name, path=path,
                              nfsShareParameters=share_param)
        resp = sr.modify_fs(nfsShareCreate=[param])
        resp.raise_if_err()
        return UnityNfsShareList(cli=cli, name=name).first_item

    @classmethod
    def create_from_snap(cls, cli, snap, name, path=None, is_read_only=None,
                         default_access=None):
        snap_clz = storops.unity.resource.snap.UnitySnap
        snap = snap_clz.get(cli, snap)
        NFSShareDefaultAccessEnum.verify(default_access)

        if path is None:
            path = '/'

        resp = cli.post(cls().resource_class,
                        snap=snap,
                        path=path,
                        name=name,
                        isReadOnly=is_read_only,
                        defaultAccess=default_access)
        resp.raise_if_err()
        return cls(_id=resp.resource_id, cli=cli)

    def delete(self, async_mode=False):
        if self.type == NFSTypeEnum.NFS_SNAPSHOT:
            resp = super(UnityNfsShare, self).delete(async_mode=async_mode)
        else:
            fs = self.filesystem.verify()
            sr = fs.storage_resource
            param = self._cli.make_body(nfsShare=self)
            resp = sr.modify_fs(async_mode=async_mode, nfsShareDelete=[param])
        resp.raise_if_err()
        return resp

    def _get_hosts(self, hosts, force_create_host=False):
        if not isinstance(hosts, (tuple, list, set, UnityResourceList)):
            hosts = [hosts]
        host_clz = storops.unity.resource.host.UnityHost
        ret = []
        for item in hosts:
            host = host_clz.get_host(self._cli, item, force_create_host,
                                     tenant=self.tenant)
            if host is not None:
                ret.append(host)
        if hosts and len(ret) == 0:
            raise UnityHostNotFoundException()
        return ret

    @property
    def host_config(self):
        # host config must be up-to-date for each call!
        self.update()
        return UnityNfsHostConfig(nfs_share=self)

    def allow_root_access(self, hosts, force_create_host=False):
        hosts = self._get_hosts(hosts, force_create_host)
        config = self.host_config.allow_root(*hosts)
        return self.modify(host_config=config)

    def allow_read_only_access(self, hosts, force_create_host=False):
        hosts = self._get_hosts(hosts, force_create_host)
        config = self.host_config.allow_ro(*hosts)
        return self.modify(host_config=config)

    def allow_read_write_access(self, hosts, force_create_host=False):
        hosts = self._get_hosts(hosts, force_create_host)
        config = self.host_config.allow_rw(*hosts)
        return self.modify(host_config=config)

    def deny_access(self, hosts, force_create_host=False):
        hosts = self._get_hosts(hosts, force_create_host)
        config = self.host_config.deny_access(*hosts)
        return self.modify(host_config=config)

    def delete_access(self, hosts):
        hosts = self._get_hosts(hosts)
        config = self.host_config.delete_access(*hosts)
        return self.modify(host_config=config)

    def clear_access(self, white_list=None, force_create_host=False):
        if white_list is not None:
            white_list = self._get_hosts(white_list, force_create_host)
        else:
            white_list = []
        config = self.host_config.clear_all(*white_list)
        return self.modify(host_config=config)

    def modify(self,
               default_access=None,
               min_security=None,
               no_access_hosts=None,
               read_only_hosts=None,
               read_write_hosts=None,
               root_access_hosts=None,
               host_config=None):
        if host_config is not None:
            no_access_hosts = host_config.no_access
            root_access_hosts = host_config.root
            read_only_hosts = host_config.ro
            read_write_hosts = host_config.rw

        NFSShareDefaultAccessEnum.verify(default_access)
        NFSShareSecurityEnum.verify(min_security)
        clz = storops.unity.resource.host.UnityHostList
        no_access_hosts = clz.get_list(self._cli, no_access_hosts)
        read_only_hosts = clz.get_list(self._cli, read_only_hosts)
        read_write_hosts = clz.get_list(self._cli, read_write_hosts)
        root_access_hosts = clz.get_list(self._cli, root_access_hosts)

        nfs_share_param = self._cli.make_body(
            allow_empty=True,
            defaultAccess=default_access,
            minSecurity=min_security,
            noAccessHosts=no_access_hosts,
            readOnlyHosts=read_only_hosts,
            readWriteHosts=read_write_hosts,
            rootAccessHosts=root_access_hosts)

        if nfs_share_param:
            # different api for different type of share
            if self.type == NFSTypeEnum.NFS_SHARE:
                resp = self._modify_fs_share(nfs_share_param)
            elif self.type == NFSTypeEnum.NFS_SNAPSHOT:
                resp = self._modify_snap_share(default_access,
                                               min_security,
                                               nfs_share_param)
            else:
                raise UnityShareTypeNotSupportAccessControlError()
        else:
            resp = RestResponse('', self._cli)
        resp.raise_if_err()
        return resp

    def _modify_snap_share(self, default_access, min_security,
                           nfs_share_param):
        return self.action('modify',
                           defaultAccess=default_access,
                           minSecurity=min_security,
                           **nfs_share_param)

    def _modify_fs_share(self, nfs_share_param):
        sr = self.storage_resource
        if sr is None:
            raise ValueError('storage resource for share {} not found.'
                             .format(self.name))
        nfs_share = self._cli.make_body(
            allow_empty=True,
            nfsShare=self,
            nfsShareParameters=nfs_share_param)
        param = self._cli.make_body(
            allow_empty=True,
            nfsShareModify=[nfs_share])
        return sr.modify_fs(**param)

    @property
    @instance_cache
    def storage_resource(self):
        fs = self.filesystem
        if fs is not None:
            ret = fs.storage_resource
        else:
            ret = None
        return ret

    @property
    @instance_cache
    def tenant(self):
        if self.filesystem is not None:
            server = self.filesystem.nas_server
            if server is not None:
                return server.tenant
        return None

    def create_snap(self, name=None, fs_access_type=None):
        if fs_access_type is None:
            fs_access_type = FilesystemSnapAccessTypeEnum.PROTOCOL

        if self.type == NFSTypeEnum.NFS_SHARE:
            ret = self.filesystem.create_snap(
                name=name, fs_access_type=fs_access_type)
        elif self.type == NFSTypeEnum.NFS_SNAPSHOT:
            ret = self.snap.copy(copy_name=name)
        else:
            raise UnityCreateSnapError('do not know how to create snap for '
                                       'nfs share {}, type {}.'
                                       .format(self.name, self.type))
        return ret


class UnityNfsShareList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityNfsShare
